import shutil
import tempfile

from unittest import TestCase
from mock import MagicMock
from autosubmit.job.job_packager import JobPackager
from autosubmit.job.job_packages import JobPackageVertical
from autosubmit.job.job import Job
from autosubmit.job.job_list import JobList
from autosubmit.job.job_dict import DicJobs
from autosubmit.job.job_utils import Dependency
from autosubmitconfigparser.config.yamlparser import YAMLParserFactory
from autosubmit.job.job_list_persistence import JobListPersistenceDb
from autosubmit.job.job_common import Status
from random import randrange
from collections import OrderedDict


class TestWrappers(TestCase):

    @classmethod
    def setUpClass(cls):
        # set up different unused_figs to be used in the test methods
        cls.workflows = dict()
        cls.workflows['basic'] = dict()
        cls.workflows['synchronize_date'] = dict()
        cls.workflows['synchronize_member'] = dict()
        cls.workflows['running_member'] = dict()
        cls.workflows['running_date'] = dict()
        cls.workflows['running_once'] = dict()

        cls.workflows['basic']['sections'] = OrderedDict()
        cls.workflows['basic']['sections']["s1"] = dict()
        cls.workflows['basic']['sections']["s1"]["RUNNING"] = "member"
        cls.workflows['basic']['sections']["s1"]["WALLCLOCK"] = '00:50'

        cls.workflows['basic']['sections']["s2"] = dict()
        cls.workflows['basic']['sections']["s2"]["RUNNING"] = "chunk"
        cls.workflows['basic']['sections']["s2"]["WALLCLOCK"] = '00:10'
        cls.workflows['basic']['sections']["s2"]["DEPENDENCIES"] = "s1 s2-1"

        cls.workflows['basic']['sections']["s3"] = dict()
        cls.workflows['basic']['sections']["s3"]["RUNNING"] = "chunk"
        cls.workflows['basic']['sections']["s3"]["WALLCLOCK"] = '00:20'
        cls.workflows['basic']['sections']["s3"]["DEPENDENCIES"] = "s2"

        cls.workflows['basic']['sections']["s4"] = dict()
        cls.workflows['basic']['sections']["s4"]["RUNNING"] = "chunk"
        cls.workflows['basic']['sections']["s4"]["WALLCLOCK"] = '00:30'
        cls.workflows['basic']['sections']["s4"]["DEPENDENCIES"] = "s3"

        cls.workflows['synchronize_date']['sections'] = OrderedDict()
        cls.workflows['synchronize_date']['sections']["s1"] = dict()
        cls.workflows['synchronize_date']['sections']["s1"]["RUNNING"] = "member"
        cls.workflows['synchronize_date']['sections']["s1"]["WALLCLOCK"] = '00:50'

        cls.workflows['synchronize_date']['sections']["s2"] = dict()
        cls.workflows['synchronize_date']['sections']["s2"]["RUNNING"] = "chunk"
        cls.workflows['synchronize_date']['sections']["s2"]["WALLCLOCK"] = '00:10'
        cls.workflows['synchronize_date']['sections']["s2"]["DEPENDENCIES"] = "s1 s2-1"

        cls.workflows['synchronize_date']['sections']["s3"] = dict()
        cls.workflows['synchronize_date']['sections']["s3"]["RUNNING"] = "chunk"
        cls.workflows['synchronize_date']['sections']["s3"]["WALLCLOCK"] = '00:20'
        cls.workflows['synchronize_date']['sections']["s3"]["DEPENDENCIES"] = "s2"

        cls.workflows['synchronize_date']['sections']["s4"] = dict()
        cls.workflows['synchronize_date']['sections']["s4"]["RUNNING"] = "chunk"
        cls.workflows['synchronize_date']['sections']["s4"]["WALLCLOCK"] = '00:30'
        cls.workflows['synchronize_date']['sections']["s4"]["DEPENDENCIES"] = "s3"

        cls.workflows['synchronize_date']['sections']["s5"] = dict()
        cls.workflows['synchronize_date']['sections']["s5"]["RUNNING"] = "chunk"
        cls.workflows['synchronize_date']['sections']["s5"]["SYNCHRONIZE"] = "date"
        cls.workflows['synchronize_date']['sections']["s5"]["WALLCLOCK"] = '00:30'
        cls.workflows['synchronize_date']['sections']["s5"]["DEPENDENCIES"] = "s2"

        cls.workflows['synchronize_member']['sections'] = OrderedDict()
        cls.workflows['synchronize_member']['sections']["s1"] = dict()
        cls.workflows['synchronize_member']['sections']["s1"]["RUNNING"] = "member"
        cls.workflows['synchronize_member']['sections']["s1"]["WALLCLOCK"] = '00:50'

        cls.workflows['synchronize_member']['sections']["s2"] = dict()
        cls.workflows['synchronize_member']['sections']["s2"]["RUNNING"] = "chunk"
        cls.workflows['synchronize_member']['sections']["s2"]["WALLCLOCK"] = '00:10'
        cls.workflows['synchronize_member']['sections']["s2"]["DEPENDENCIES"] = "s1 s2-1"

        cls.workflows['synchronize_member']['sections']["s3"] = dict()
        cls.workflows['synchronize_member']['sections']["s3"]["RUNNING"] = "chunk"
        cls.workflows['synchronize_member']['sections']["s3"]["WALLCLOCK"] = '00:20'
        cls.workflows['synchronize_member']['sections']["s3"]["DEPENDENCIES"] = "s2"

        cls.workflows['synchronize_member']['sections']["s4"] = dict()
        cls.workflows['synchronize_member']['sections']["s4"]["RUNNING"] = "chunk"
        cls.workflows['synchronize_member']['sections']["s4"]["WALLCLOCK"] = '00:30'
        cls.workflows['synchronize_member']['sections']["s4"]["DEPENDENCIES"] = "s3"

        cls.workflows['synchronize_member']['sections']["s5"] = dict()
        cls.workflows['synchronize_member']['sections']["s5"]["RUNNING"] = "chunk"
        cls.workflows['synchronize_member']['sections']["s5"]["SYNCHRONIZE"] = "member"
        cls.workflows['synchronize_member']['sections']["s5"]["WALLCLOCK"] = '00:30'
        cls.workflows['synchronize_member']['sections']["s5"]["DEPENDENCIES"] = "s2"

        cls.workflows['running_date']['sections'] = OrderedDict()
        cls.workflows['running_date']['sections']["s1"] = dict()
        cls.workflows['running_date']['sections']["s1"]["RUNNING"] = "member"
        cls.workflows['running_date']['sections']["s1"]["WALLCLOCK"] = '00:50'

        cls.workflows['running_date']['sections']["s2"] = dict()
        cls.workflows['running_date']['sections']["s2"]["RUNNING"] = "chunk"
        cls.workflows['running_date']['sections']["s2"]["WALLCLOCK"] = '00:10'
        cls.workflows['running_date']['sections']["s2"]["DEPENDENCIES"] = "s1 s2-1"

        cls.workflows['running_date']['sections']["s3"] = dict()
        cls.workflows['running_date']['sections']["s3"]["RUNNING"] = "chunk"
        cls.workflows['running_date']['sections']["s3"]["WALLCLOCK"] = '00:20'
        cls.workflows['running_date']['sections']["s3"]["DEPENDENCIES"] = "s2"

        cls.workflows['running_date']['sections']["s4"] = dict()
        cls.workflows['running_date']['sections']["s4"]["RUNNING"] = "chunk"
        cls.workflows['running_date']['sections']["s4"]["WALLCLOCK"] = '00:30'
        cls.workflows['running_date']['sections']["s4"]["DEPENDENCIES"] = "s3"

        cls.workflows['running_date']['sections']["s5"] = dict()
        cls.workflows['running_date']['sections']["s5"]["RUNNING"] = "date"
        cls.workflows['running_date']['sections']["s5"]["WALLCLOCK"] = '00:30'
        cls.workflows['running_date']['sections']["s5"]["DEPENDENCIES"] = "s2"

        cls.workflows['running_once']['sections'] = OrderedDict()
        cls.workflows['running_once']['sections']["s1"] = dict()
        cls.workflows['running_once']['sections']["s1"]["RUNNING"] = "member"
        cls.workflows['running_once']['sections']["s1"]["WALLCLOCK"] = '00:50'

        cls.workflows['running_once']['sections']["s2"] = dict()
        cls.workflows['running_once']['sections']["s2"]["RUNNING"] = "chunk"
        cls.workflows['running_once']['sections']["s2"]["WALLCLOCK"] = '00:10'
        cls.workflows['running_once']['sections']["s2"]["DEPENDENCIES"] = "s1 s2-1"

        cls.workflows['running_once']['sections']["s3"] = dict()
        cls.workflows['running_once']['sections']["s3"]["RUNNING"] = "chunk"
        cls.workflows['running_once']['sections']["s3"]["WALLCLOCK"] = '00:20'
        cls.workflows['running_once']['sections']["s3"]["DEPENDENCIES"] = "s2"

        cls.workflows['running_once']['sections']["s4"] = dict()
        cls.workflows['running_once']['sections']["s4"]["RUNNING"] = "chunk"
        cls.workflows['running_once']['sections']["s4"]["WALLCLOCK"] = '00:30'
        cls.workflows['running_once']['sections']["s4"]["DEPENDENCIES"] = "s3"

        cls.workflows['running_once']['sections']["s5"] = dict()
        cls.workflows['running_once']['sections']["s5"]["RUNNING"] = "once"
        cls.workflows['running_once']['sections']["s5"]["WALLCLOCK"] = '00:30'
        cls.workflows['running_once']['sections']["s5"]["DEPENDENCIES"] = "s2"

    def setUp(self):
        self.experiment_id = 'random-id'
        self._wrapper_factory = MagicMock()

        self.config = FakeBasicConfig
        self._platform = MagicMock()
        self.as_conf = MagicMock()
        self.as_conf.experiment_data = dict()
        self.as_conf.experiment_data["JOBS"] = dict()
        self.as_conf.jobs_data = self.as_conf.experiment_data["JOBS"]

        self.as_conf.experiment_data["PLATFORMS"] = dict()
        self.as_conf.experiment_data["WRAPPERS"] = dict()
        self.temp_directory = tempfile.mkdtemp()
        self.job_list = JobList(self.experiment_id, self.config, YAMLParserFactory(),
                                JobListPersistenceDb(self.temp_directory, 'db'),self.as_conf)
        self.parser_mock = MagicMock(spec='SafeConfigParser')

        self._platform.max_waiting_jobs = 100
        self._platform.total_jobs = 100
        self.config.get_wrapper_type = MagicMock(return_value='vertical')
        self.config.get_wrapper_export = MagicMock(return_value='')
        self.config.get_wrapper_jobs = MagicMock(return_value='None')
        self.config.get_wrapper_method = MagicMock(return_value='ASThread')
        self.config.get_wrapper_queue = MagicMock(return_value='debug')
        self.config.get_wrapper_policy = MagicMock(return_value='flexible')
        self.config.get_extensible_wallclock = MagicMock(return_value=0)
        self.config.get_retrials = MagicMock(return_value=0)
        options = {
            'TYPE': "vertical",
            'JOBS_IN_WRAPPER': "None",
            'EXPORT': "none",
            'METHOD': "ASThread",
            'QUEUE': "debug",
            'POLICY': "flexible",
            'RETRIALS': 0,
            'EXTEND_WALLCLOCK': 0
        }
        self.as_conf.experiment_data["WRAPPERS"]["WRAPPERS"] = options
        self.as_conf.experiment_data["WRAPPERS"]["CURRENT_WRAPPER"] = options
        self._wrapper_factory.as_conf = self.as_conf
        self.job_packager = JobPackager(
            self.as_conf, self._platform, self.job_list)
        self.job_list._ordered_jobs_by_date_member["WRAPPERS"] = dict()

    def tearDown(self) -> None:
        shutil.rmtree(self.temp_directory)

    ### ONE SECTION WRAPPER ###
    def test_returned_packages(self):
        self.current_wrapper_section = {}
        date_list = ["d1", "d2"]
        member_list = ["m1", "m2"]
        chunk_list = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
        for section,s_value in self.workflows['basic']['sections'].items():
            self.as_conf.jobs_data[section] = s_value
        self._createDummyJobs(
            self.workflows['basic'], date_list, member_list, chunk_list)

        self.job_list.get_job_by_name(
            'expid_d1_m1_s1').status = Status.COMPLETED
        self.job_list.get_job_by_name(
            'expid_d1_m2_s1').status = Status.COMPLETED

        self.job_list.get_job_by_name('expid_d1_m1_1_s2').status = Status.READY
        self.job_list.get_job_by_name('expid_d1_m2_1_s2').status = Status.READY

        max_jobs = 20
        max_wrapped_jobs = 20
        max_wallclock = '10:00'

        d1_m1_1_s2 = self.job_list.get_job_by_name('expid_d1_m1_1_s2')
        d1_m1_2_s2 = self.job_list.get_job_by_name('expid_d1_m1_2_s2')
        d1_m1_3_s2 = self.job_list.get_job_by_name('expid_d1_m1_3_s2')
        d1_m1_4_s2 = self.job_list.get_job_by_name('expid_d1_m1_4_s2')
        d1_m1_5_s2 = self.job_list.get_job_by_name('expid_d1_m1_5_s2')
        d1_m1_6_s2 = self.job_list.get_job_by_name('expid_d1_m1_6_s2')
        d1_m1_7_s2 = self.job_list.get_job_by_name('expid_d1_m1_7_s2')
        d1_m1_8_s2 = self.job_list.get_job_by_name('expid_d1_m1_8_s2')
        d1_m1_9_s2 = self.job_list.get_job_by_name('expid_d1_m1_9_s2')
        d1_m1_10_s2 = self.job_list.get_job_by_name('expid_d1_m1_10_s2')

        d1_m2_1_s2 = self.job_list.get_job_by_name('expid_d1_m2_1_s2')
        d1_m2_2_s2 = self.job_list.get_job_by_name('expid_d1_m2_2_s2')
        d1_m2_3_s2 = self.job_list.get_job_by_name('expid_d1_m2_3_s2')
        d1_m2_4_s2 = self.job_list.get_job_by_name('expid_d1_m2_4_s2')
        d1_m2_5_s2 = self.job_list.get_job_by_name('expid_d1_m2_5_s2')
        d1_m2_6_s2 = self.job_list.get_job_by_name('expid_d1_m2_6_s2')
        d1_m2_7_s2 = self.job_list.get_job_by_name('expid_d1_m2_7_s2')
        d1_m2_8_s2 = self.job_list.get_job_by_name('expid_d1_m2_8_s2')
        d1_m2_9_s2 = self.job_list.get_job_by_name('expid_d1_m2_9_s2')
        d1_m2_10_s2 = self.job_list.get_job_by_name('expid_d1_m2_10_s2')
        self.job_list._ordered_jobs_by_date_member["WRAPPERS"]["d1"] = dict()
        self.job_list._ordered_jobs_by_date_member["WRAPPERS"]["d1"]["m1"] = [d1_m1_1_s2, d1_m1_2_s2, d1_m1_3_s2, d1_m1_4_s2, d1_m1_5_s2, d1_m1_6_s2, d1_m1_7_s2, d1_m1_8_s2, d1_m1_9_s2, d1_m1_10_s2]

        self.job_list._ordered_jobs_by_date_member["WRAPPERS"]["d1"]["m2"] = [d1_m2_1_s2, d1_m2_2_s2, d1_m2_3_s2, d1_m2_4_s2, d1_m2_5_s2, d1_m2_6_s2, d1_m2_7_s2, d1_m2_8_s2, d1_m2_9_s2, d1_m2_10_s2]
        section_list = [d1_m1_1_s2, d1_m2_1_s2]
        self.job_packager.current_wrapper_section = "WRAPPERS"
        self.job_packager.max_jobs = max_jobs
        self.job_packager.retrials = 0
        self.job_packager._platform.max_wallclock = max_wallclock
        self.job_packager.wrapper_type = 'vertical'

        max_wrapped_job_by_section = {}
        max_wrapped_job_by_section["s1"] = max_wrapped_jobs
        max_wrapped_job_by_section["s2"] = max_wrapped_jobs
        max_wrapped_job_by_section["s3"] = max_wrapped_jobs
        max_wrapped_job_by_section["s4"] = max_wrapped_jobs
        wrapper_limits = dict()
        wrapper_limits["max"] = max_wrapped_jobs
        wrapper_limits["max_v"] = max_wrapped_jobs
        wrapper_limits["max_h"] = max_wrapped_jobs
        wrapper_limits["min"] = 2
        wrapper_limits["min_v"] = 2
        wrapper_limits["min_h"] = 2
        wrapper_limits["max_by_section"] = max_wrapped_job_by_section
        returned_packages = self.job_packager._build_vertical_packages(
            section_list, wrapper_limits)

        package_m1_s2 = [d1_m1_1_s2, d1_m1_2_s2, d1_m1_3_s2, d1_m1_4_s2, d1_m1_5_s2, d1_m1_6_s2, d1_m1_7_s2, d1_m1_8_s2,
                         d1_m1_9_s2, d1_m1_10_s2]
        package_m2_s2 = [d1_m2_1_s2, d1_m2_2_s2, d1_m2_3_s2, d1_m2_4_s2, d1_m2_5_s2, d1_m2_6_s2, d1_m2_7_s2, d1_m2_8_s2,
                         d1_m2_9_s2, d1_m2_10_s2]

        packages = [JobPackageVertical(package_m1_s2,configuration=self.as_conf), JobPackageVertical(package_m2_s2,configuration=self.as_conf)]

        # returned_packages = returned_packages[]
        for i in range(0, len(returned_packages)):
            self.assertListEqual(returned_packages[i]._jobs, packages[i]._jobs)

    def test_returned_packages_max_jobs(self):
        date_list = ["d1", "d2"]
        member_list = ["m1", "m2"]
        chunk_list = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]

        for section,s_value in self.workflows['basic']['sections'].items():
            self.as_conf.jobs_data[section] = s_value
        self._createDummyJobs(
            self.workflows['basic'], date_list, member_list, chunk_list)

        self.job_list.get_job_by_name(
            'expid_d1_m1_s1').status = Status.COMPLETED
        self.job_list.get_job_by_name(
            'expid_d1_m2_s1').status = Status.COMPLETED

        self.job_list.get_job_by_name('expid_d1_m1_1_s2').status = Status.READY
        self.job_list.get_job_by_name('expid_d1_m2_1_s2').status = Status.READY

        max_jobs = 12
        max_wrapped_jobs = 10
        max_wallclock = '10:00'

        d1_m1_1_s2 = self.job_list.get_job_by_name('expid_d1_m1_1_s2')
        d1_m1_2_s2 = self.job_list.get_job_by_name('expid_d1_m1_2_s2')
        d1_m1_3_s2 = self.job_list.get_job_by_name('expid_d1_m1_3_s2')
        d1_m1_4_s2 = self.job_list.get_job_by_name('expid_d1_m1_4_s2')
        d1_m1_5_s2 = self.job_list.get_job_by_name('expid_d1_m1_5_s2')
        d1_m1_6_s2 = self.job_list.get_job_by_name('expid_d1_m1_6_s2')
        d1_m1_7_s2 = self.job_list.get_job_by_name('expid_d1_m1_7_s2')
        d1_m1_8_s2 = self.job_list.get_job_by_name('expid_d1_m1_8_s2')
        d1_m1_9_s2 = self.job_list.get_job_by_name('expid_d1_m1_9_s2')
        d1_m1_10_s2 = self.job_list.get_job_by_name('expid_d1_m1_10_s2')

        d1_m2_1_s2 = self.job_list.get_job_by_name('expid_d1_m2_1_s2')
        d1_m2_2_s2 = self.job_list.get_job_by_name('expid_d1_m2_2_s2')
        d1_m2_3_s2 = self.job_list.get_job_by_name('expid_d1_m2_3_s2')
        d1_m2_4_s2 = self.job_list.get_job_by_name('expid_d1_m2_4_s2')
        d1_m2_5_s2 = self.job_list.get_job_by_name('expid_d1_m2_5_s2')
        d1_m2_6_s2 = self.job_list.get_job_by_name('expid_d1_m2_6_s2')
        d1_m2_7_s2 = self.job_list.get_job_by_name('expid_d1_m2_7_s2')
        d1_m2_8_s2 = self.job_list.get_job_by_name('expid_d1_m2_8_s2')
        d1_m2_9_s2 = self.job_list.get_job_by_name('expid_d1_m2_9_s2')
        d1_m2_10_s2 = self.job_list.get_job_by_name('expid_d1_m2_10_s2')
        self.job_list._ordered_jobs_by_date_member["WRAPPERS"]["d1"] = dict()
        self.job_list._ordered_jobs_by_date_member["WRAPPERS"]["d1"]["m1"] = [d1_m1_1_s2, d1_m1_2_s2, d1_m1_3_s2, d1_m1_4_s2, d1_m1_5_s2, d1_m1_6_s2, d1_m1_7_s2, d1_m1_8_s2, d1_m1_9_s2, d1_m1_10_s2]

        self.job_list._ordered_jobs_by_date_member["WRAPPERS"]["d1"]["m2"] = [d1_m2_1_s2, d1_m2_2_s2, d1_m2_3_s2, d1_m2_4_s2, d1_m2_5_s2, d1_m2_6_s2, d1_m2_7_s2, d1_m2_8_s2, d1_m2_9_s2, d1_m2_10_s2]

        section_list = [d1_m1_1_s2, d1_m2_1_s2]

        self.job_packager.max_jobs = max_jobs
        self.job_packager._platform.max_wallclock = max_wallclock
        self.job_packager.wrapper_type = 'vertical'
        self.job_packager.retrials = 0
        max_wrapped_job_by_section = {}
        max_wrapped_job_by_section["s1"] = max_wrapped_jobs
        max_wrapped_job_by_section["s2"] = max_wrapped_jobs
        max_wrapped_job_by_section["s3"] = max_wrapped_jobs
        max_wrapped_job_by_section["s4"] = max_wrapped_jobs
        wrapper_limits = dict()
        wrapper_limits["max"] = max_wrapped_jobs
        wrapper_limits["max_v"] = max_wrapped_jobs
        wrapper_limits["max_h"] = max_wrapped_jobs
        wrapper_limits["min"] = 2
        wrapper_limits["min_v"] = 2
        wrapper_limits["min_h"] = 2
        wrapper_limits["max_by_section"] = max_wrapped_job_by_section
        returned_packages = self.job_packager._build_vertical_packages(
            section_list, wrapper_limits)

        package_m1_s2 = [d1_m1_1_s2, d1_m1_2_s2, d1_m1_3_s2, d1_m1_4_s2, d1_m1_5_s2, d1_m1_6_s2, d1_m1_7_s2, d1_m1_8_s2,
                         d1_m1_9_s2, d1_m1_10_s2]
        package_m2_s2 = [d1_m2_1_s2, d1_m2_2_s2, d1_m2_3_s2, d1_m2_4_s2, d1_m2_5_s2, d1_m2_6_s2, d1_m2_7_s2, d1_m2_8_s2,
                         d1_m2_9_s2, d1_m2_10_s2]

        packages = [JobPackageVertical(
            package_m1_s2,configuration=self.as_conf), JobPackageVertical(package_m2_s2,configuration=self.as_conf)]

        for i in range(0, len(returned_packages)):
            self.assertListEqual(returned_packages[i]._jobs, packages[i]._jobs)

    def test_returned_packages_max_wrapped_jobs(self):
        date_list = ["d1", "d2"]
        member_list = ["m1", "m2"]
        chunk_list = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
        for section,s_value in self.workflows['basic']['sections'].items():
            self.as_conf.jobs_data[section] = s_value
        self._createDummyJobs(
            self.workflows['basic'], date_list, member_list, chunk_list)

        self.job_list.get_job_by_name(
            'expid_d1_m1_s1').status = Status.COMPLETED
        self.job_list.get_job_by_name(
            'expid_d1_m2_s1').status = Status.COMPLETED

        self.job_list.get_job_by_name('expid_d1_m1_1_s2').status = Status.READY
        self.job_list.get_job_by_name('expid_d1_m2_1_s2').status = Status.READY

        max_jobs = 20
        max_wrapped_jobs = 5
        max_wallclock = '10:00'

        d1_m1_1_s2 = self.job_list.get_job_by_name('expid_d1_m1_1_s2')
        d1_m1_2_s2 = self.job_list.get_job_by_name('expid_d1_m1_2_s2')
        d1_m1_3_s2 = self.job_list.get_job_by_name('expid_d1_m1_3_s2')
        d1_m1_4_s2 = self.job_list.get_job_by_name('expid_d1_m1_4_s2')
        d1_m1_5_s2 = self.job_list.get_job_by_name('expid_d1_m1_5_s2')

        d1_m2_1_s2 = self.job_list.get_job_by_name('expid_d1_m2_1_s2')
        d1_m2_2_s2 = self.job_list.get_job_by_name('expid_d1_m2_2_s2')
        d1_m2_3_s2 = self.job_list.get_job_by_name('expid_d1_m2_3_s2')
        d1_m2_4_s2 = self.job_list.get_job_by_name('expid_d1_m2_4_s2')
        d1_m2_5_s2 = self.job_list.get_job_by_name('expid_d1_m2_5_s2')
        self.job_list._ordered_jobs_by_date_member["WRAPPERS"]["d1"] = dict()
        self.job_list._ordered_jobs_by_date_member["WRAPPERS"]["d1"]["m1"] = [d1_m1_1_s2, d1_m1_2_s2, d1_m1_3_s2, d1_m1_4_s2, d1_m1_5_s2]

        self.job_list._ordered_jobs_by_date_member["WRAPPERS"]["d1"]["m2"] = [d1_m2_1_s2, d1_m2_2_s2, d1_m2_3_s2, d1_m2_4_s2, d1_m2_5_s2]

        section_list = [d1_m1_1_s2, d1_m2_1_s2]

        self.job_packager.max_jobs = max_jobs
        self.job_packager._platform.max_wallclock = max_wallclock
        self.job_packager.wrapper_type = 'vertical'
        self.job_packager.retrials = 0
        max_wrapped_job_by_section = {}
        max_wrapped_job_by_section["s1"] = max_wrapped_jobs
        max_wrapped_job_by_section["s2"] = max_wrapped_jobs
        max_wrapped_job_by_section["s3"] = max_wrapped_jobs
        max_wrapped_job_by_section["s4"] = max_wrapped_jobs
        wrapper_limits = dict()
        wrapper_limits["max"] = max_wrapped_jobs
        wrapper_limits["max_v"] = max_wrapped_jobs
        wrapper_limits["max_h"] = max_wrapped_jobs
        wrapper_limits["min"] = 2
        wrapper_limits["min_v"] = 2
        wrapper_limits["min_h"] = 2
        wrapper_limits["max_by_section"] = max_wrapped_job_by_section
        returned_packages = self.job_packager._build_vertical_packages(
            section_list, wrapper_limits)

        package_m1_s2 = [d1_m1_1_s2, d1_m1_2_s2,
                         d1_m1_3_s2, d1_m1_4_s2, d1_m1_5_s2]
        package_m2_s2 = [d1_m2_1_s2, d1_m2_2_s2,
                         d1_m2_3_s2, d1_m2_4_s2, d1_m2_5_s2]

        packages = [JobPackageVertical(
            package_m1_s2,configuration=self.as_conf), JobPackageVertical(package_m2_s2,configuration=self.as_conf)]

        #returned_packages = returned_packages[0]
        for i in range(0, len(returned_packages)):
            self.assertListEqual(returned_packages[i]._jobs, packages[i]._jobs)

    def test_returned_packages_max_wallclock(self):
        date_list = ["d1", "d2"]
        member_list = ["m1", "m2"]
        chunk_list = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
        for section,s_value in self.workflows['basic']['sections'].items():
            self.as_conf.jobs_data[section] = s_value
        self._createDummyJobs(
            self.workflows['basic'], date_list, member_list, chunk_list)

        self.job_list.get_job_by_name(
            'expid_d1_m1_s1').status = Status.COMPLETED
        self.job_list.get_job_by_name(
            'expid_d1_m2_s1').status = Status.COMPLETED

        self.job_list.get_job_by_name('expid_d1_m1_1_s2').status = Status.READY
        self.job_list.get_job_by_name('expid_d1_m2_1_s2').status = Status.READY

        max_jobs = 20
        max_wrapped_jobs = 15
        max_wallclock = '00:50'

        d1_m1_1_s2 = self.job_list.get_job_by_name('expid_d1_m1_1_s2')
        d1_m1_2_s2 = self.job_list.get_job_by_name('expid_d1_m1_2_s2')
        d1_m1_3_s2 = self.job_list.get_job_by_name('expid_d1_m1_3_s2')
        d1_m1_4_s2 = self.job_list.get_job_by_name('expid_d1_m1_4_s2')
        d1_m1_5_s2 = self.job_list.get_job_by_name('expid_d1_m1_5_s2')

        d1_m2_1_s2 = self.job_list.get_job_by_name('expid_d1_m2_1_s2')
        d1_m2_2_s2 = self.job_list.get_job_by_name('expid_d1_m2_2_s2')
        d1_m2_3_s2 = self.job_list.get_job_by_name('expid_d1_m2_3_s2')
        d1_m2_4_s2 = self.job_list.get_job_by_name('expid_d1_m2_4_s2')
        d1_m2_5_s2 = self.job_list.get_job_by_name('expid_d1_m2_5_s2')
        self.job_list._ordered_jobs_by_date_member["WRAPPERS"]["d1"] = dict()
        self.job_list._ordered_jobs_by_date_member["WRAPPERS"]["d1"]["m1"] = [d1_m1_1_s2, d1_m1_2_s2, d1_m1_3_s2, d1_m1_4_s2, d1_m1_5_s2]

        self.job_list._ordered_jobs_by_date_member["WRAPPERS"]["d1"]["m2"] = [d1_m2_1_s2, d1_m2_2_s2, d1_m2_3_s2, d1_m2_4_s2, d1_m2_5_s2]

        section_list = [d1_m1_1_s2, d1_m2_1_s2]

        self.job_packager.max_jobs = max_jobs
        self.job_packager._platform.max_wallclock = max_wallclock
        self.job_packager.wrapper_type = 'vertical'
        self.job_packager.retrials = 0
        max_wrapped_job_by_section = {}
        max_wrapped_job_by_section["s1"] = max_wrapped_jobs
        max_wrapped_job_by_section["s2"] = max_wrapped_jobs
        max_wrapped_job_by_section["s3"] = max_wrapped_jobs
        max_wrapped_job_by_section["s4"] = max_wrapped_jobs
        wrapper_limits = dict()
        wrapper_limits["max"] = max_wrapped_jobs
        wrapper_limits["max_v"] = max_wrapped_jobs
        wrapper_limits["max_h"] = max_wrapped_jobs
        wrapper_limits["min"] = 2
        wrapper_limits["min_v"] = 2
        wrapper_limits["min_h"] = 2
        wrapper_limits["max_by_section"] = max_wrapped_job_by_section
        returned_packages = self.job_packager._build_vertical_packages(
            section_list, wrapper_limits)

        package_m1_s2 = [d1_m1_1_s2, d1_m1_2_s2,
                         d1_m1_3_s2, d1_m1_4_s2, d1_m1_5_s2]
        package_m2_s2 = [d1_m2_1_s2, d1_m2_2_s2,
                         d1_m2_3_s2, d1_m2_4_s2, d1_m2_5_s2]

        packages = [JobPackageVertical(
            package_m1_s2,configuration=self.as_conf), JobPackageVertical(package_m2_s2,configuration=self.as_conf)]

        #returned_packages = returned_packages[0]
        for i in range(0, len(returned_packages)):
            self.assertListEqual(returned_packages[i]._jobs, packages[i]._jobs)

    def test_returned_packages_section_not_self_dependent(self):
        date_list = ["d1", "d2"]
        member_list = ["m1", "m2"]
        chunk_list = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
        for section,s_value in self.workflows['basic']['sections'].items():
            self.as_conf.jobs_data[section] = s_value
        self._createDummyJobs(
            self.workflows['basic'], date_list, member_list, chunk_list)

        self.job_list.get_job_by_name(
            'expid_d1_m1_s1').status = Status.COMPLETED
        self.job_list.get_job_by_name(
            'expid_d1_m2_s1').status = Status.COMPLETED
        self.job_list.get_job_by_name(
            'expid_d1_m1_1_s2').status = Status.COMPLETED
        self.job_list.get_job_by_name(
            'expid_d1_m2_1_s2').status = Status.COMPLETED

        self.job_list.get_job_by_name('expid_d1_m1_1_s3').status = Status.READY
        self.job_list.get_job_by_name('expid_d1_m2_1_s3').status = Status.READY

        max_jobs = 20
        max_wrapped_jobs = 20
        max_wallclock = '10:00'

        d1_m1_1_s3 = self.job_list.get_job_by_name('expid_d1_m1_1_s3')
        d1_m2_1_s3 = self.job_list.get_job_by_name('expid_d1_m2_1_s3')
        self.job_list._ordered_jobs_by_date_member["WRAPPERS"]["d1"] = dict()
        self.job_list._ordered_jobs_by_date_member["WRAPPERS"]["d1"]["m1"] = [d1_m1_1_s3]

        self.job_list._ordered_jobs_by_date_member["WRAPPERS"]["d1"]["m2"] = [d1_m2_1_s3]

        section_list = [d1_m1_1_s3, d1_m2_1_s3]

        self.job_packager.max_jobs = max_jobs
        self.job_packager._platform.max_wallclock = max_wallclock
        self.job_packager.wrapper_type = 'vertical'
        self.job_packager.retrials = 0
        max_wrapped_job_by_section = {}
        max_wrapped_job_by_section["s1"] = max_wrapped_jobs
        max_wrapped_job_by_section["s2"] = max_wrapped_jobs
        max_wrapped_job_by_section["s3"] = max_wrapped_jobs
        max_wrapped_job_by_section["s4"] = max_wrapped_jobs
        wrapper_limits = dict()
        wrapper_limits["max"] = max_wrapped_jobs
        wrapper_limits["max_v"] = max_wrapped_jobs
        wrapper_limits["max_h"] = max_wrapped_jobs
        wrapper_limits["min"] = 2
        wrapper_limits["min_v"] = 2
        wrapper_limits["min_h"] = 2
        wrapper_limits["max_by_section"] = max_wrapped_job_by_section
        returned_packages = self.job_packager._build_vertical_packages(
            section_list, wrapper_limits)
        self.job_list._ordered_jobs_by_date_member["WRAPPERS"]["d1"] = dict()
        self.job_list._ordered_jobs_by_date_member["WRAPPERS"]["d1"]["m1"] = [d1_m1_1_s3]

        self.job_list._ordered_jobs_by_date_member["WRAPPERS"]["d1"]["m2"] = [d1_m2_1_s3]

        package_m1_s2 = [d1_m1_1_s3]
        package_m2_s2 = [d1_m2_1_s3]

        packages = [JobPackageVertical(
            package_m1_s2,configuration=self.as_conf), JobPackageVertical(package_m2_s2,configuration=self.as_conf)]

        #returned_packages = returned_packages[0]
        for i in range(0, len(returned_packages)):
            self.assertListEqual(returned_packages[i]._jobs, packages[i]._jobs)

    ### MIXED WRAPPER ###
    def test_returned_packages_mixed_wrapper(self):
        date_list = ["d1"]
        member_list = ["m1", "m2"]
        chunk_list = [1, 2, 3, 4]
        for section,s_value in self.workflows['basic']['sections'].items():
            self.as_conf.jobs_data[section] = s_value
        self._createDummyJobs(
            self.workflows['basic'], date_list, member_list, chunk_list)

        self.job_list.get_job_by_name(
            'expid_d1_m1_s1').status = Status.COMPLETED
        self.job_list.get_job_by_name(
            'expid_d1_m2_s1').status = Status.COMPLETED

        self.job_list.get_job_by_name('expid_d1_m1_1_s2').status = Status.READY
        self.job_list.get_job_by_name('expid_d1_m2_1_s2').status = Status.READY

        wrapper_expression = "s2 s3"
        max_jobs = 18
        max_wrapped_jobs = 18
        max_wallclock = '10:00'

        d1_m1_1_s2 = self.job_list.get_job_by_name('expid_d1_m1_1_s2')
        d1_m1_2_s2 = self.job_list.get_job_by_name('expid_d1_m1_2_s2')
        d1_m1_3_s2 = self.job_list.get_job_by_name('expid_d1_m1_3_s2')
        d1_m1_4_s2 = self.job_list.get_job_by_name('expid_d1_m1_4_s2')
        d1_m2_1_s2 = self.job_list.get_job_by_name('expid_d1_m2_1_s2')
        d1_m2_2_s2 = self.job_list.get_job_by_name('expid_d1_m2_2_s2')
        d1_m2_3_s2 = self.job_list.get_job_by_name('expid_d1_m2_3_s2')
        d1_m2_4_s2 = self.job_list.get_job_by_name('expid_d1_m2_4_s2')

        d1_m1_1_s3 = self.job_list.get_job_by_name('expid_d1_m1_1_s3')
        d1_m1_2_s3 = self.job_list.get_job_by_name('expid_d1_m1_2_s3')
        d1_m1_3_s3 = self.job_list.get_job_by_name('expid_d1_m1_3_s3')
        d1_m1_4_s3 = self.job_list.get_job_by_name('expid_d1_m1_4_s3')
        d1_m2_1_s3 = self.job_list.get_job_by_name('expid_d1_m2_1_s3')
        d1_m2_2_s3 = self.job_list.get_job_by_name('expid_d1_m2_2_s3')
        d1_m2_3_s3 = self.job_list.get_job_by_name('expid_d1_m2_3_s3')
        d1_m2_4_s3 = self.job_list.get_job_by_name('expid_d1_m2_4_s3')

        self.job_list._ordered_jobs_by_date_member["WRAPPERS"]["d1"] = dict()
        self.job_list._ordered_jobs_by_date_member["WRAPPERS"]["d1"]["m1"] = [d1_m1_1_s2, d1_m1_1_s3, d1_m1_2_s2, d1_m1_2_s3,
                                                                  d1_m1_3_s2, d1_m1_3_s3, d1_m1_4_s2, d1_m1_4_s3]

        self.job_list._ordered_jobs_by_date_member["WRAPPERS"]["d1"]["m2"] = [d1_m2_1_s2, d1_m2_1_s3, d1_m2_2_s2, d1_m2_2_s3,
                                                                  d1_m2_3_s2, d1_m2_3_s3, d1_m2_4_s2, d1_m2_4_s3]

        section_list = [d1_m1_1_s2, d1_m2_1_s2]

        self.job_packager.max_jobs = max_jobs
        self.job_packager._platform.max_wallclock = max_wallclock
        self.job_packager.wrapper_type = 'vertical'
        self.job_packager.retrials = 0
        self.job_packager.jobs_in_wrapper = wrapper_expression
        max_wrapped_job_by_section = {}
        max_wrapped_job_by_section["s1"] = max_wrapped_jobs
        max_wrapped_job_by_section["s2"] = max_wrapped_jobs
        max_wrapped_job_by_section["s3"] = max_wrapped_jobs
        max_wrapped_job_by_section["s4"] = max_wrapped_jobs
        wrapper_limits = dict()
        wrapper_limits["max"] = max_wrapped_jobs
        wrapper_limits["max_v"] = max_wrapped_jobs
        wrapper_limits["max_h"] = max_wrapped_jobs
        wrapper_limits["min"] = 2
        wrapper_limits["min_v"] = 2
        wrapper_limits["min_h"] = 2
        wrapper_limits["max_by_section"] = max_wrapped_job_by_section
        returned_packages = self.job_packager._build_vertical_packages(
            section_list, wrapper_limits)

        package_m1_s2_s3 = [d1_m1_1_s2, d1_m1_1_s3, d1_m1_2_s2, d1_m1_2_s3, d1_m1_3_s2, d1_m1_3_s3, d1_m1_4_s2,
                            d1_m1_4_s3]
        package_m2_s2_s3 = [d1_m2_1_s2, d1_m2_1_s3, d1_m2_2_s2, d1_m2_2_s3, d1_m2_3_s2, d1_m2_3_s3, d1_m2_4_s2,
                            d1_m2_4_s3]

        packages = [JobPackageVertical(
            package_m1_s2_s3,configuration=self.as_conf), JobPackageVertical(package_m2_s2_s3,configuration=self.as_conf)]

        #returned_packages = returned_packages[0]
        for i in range(0, len(returned_packages)):
            self.assertListEqual(returned_packages[i]._jobs, packages[i]._jobs)

    def test_returned_packages_parent_failed_mixed_wrapper(self):
        date_list = ["d1"]
        member_list = ["m1", "m2"]
        chunk_list = [1, 2, 3, 4]
        for section,s_value in self.workflows['basic']['sections'].items():
            self.as_conf.jobs_data[section] = s_value
        self._createDummyJobs(
            self.workflows['basic'], date_list, member_list, chunk_list)

        self.job_list.get_job_by_name(
            'expid_d1_m1_s1').status = Status.COMPLETED
        self.job_list.get_job_by_name('expid_d1_m2_s1').status = Status.FAILED

        self.job_list.get_job_by_name('expid_d1_m1_1_s2').status = Status.READY

        wrapper_expression = "s2 s3"
        max_jobs = 18
        max_wrapped_jobs = 18
        max_wallclock = '10:00'

        d1_m1_1_s2 = self.job_list.get_job_by_name('expid_d1_m1_1_s2')
        d1_m1_2_s2 = self.job_list.get_job_by_name('expid_d1_m1_2_s2')
        d1_m1_3_s2 = self.job_list.get_job_by_name('expid_d1_m1_3_s2')
        d1_m1_4_s2 = self.job_list.get_job_by_name('expid_d1_m1_4_s2')
        d1_m2_1_s2 = self.job_list.get_job_by_name('expid_d1_m2_1_s2')
        d1_m2_2_s2 = self.job_list.get_job_by_name('expid_d1_m2_2_s2')
        d1_m2_3_s2 = self.job_list.get_job_by_name('expid_d1_m2_3_s2')
        d1_m2_4_s2 = self.job_list.get_job_by_name('expid_d1_m2_4_s2')

        d1_m1_1_s3 = self.job_list.get_job_by_name('expid_d1_m1_1_s3')
        d1_m1_2_s3 = self.job_list.get_job_by_name('expid_d1_m1_2_s3')
        d1_m1_3_s3 = self.job_list.get_job_by_name('expid_d1_m1_3_s3')
        d1_m1_4_s3 = self.job_list.get_job_by_name('expid_d1_m1_4_s3')
        d1_m2_1_s3 = self.job_list.get_job_by_name('expid_d1_m2_1_s3')
        d1_m2_2_s3 = self.job_list.get_job_by_name('expid_d1_m2_2_s3')
        d1_m2_3_s3 = self.job_list.get_job_by_name('expid_d1_m2_3_s3')
        d1_m2_4_s3 = self.job_list.get_job_by_name('expid_d1_m2_4_s3')
        self.job_list._ordered_jobs_by_date_member["WRAPPERS"] = dict()
        self.job_list._ordered_jobs_by_date_member["WRAPPERS"]["d1"] = dict()
        self.job_list._ordered_jobs_by_date_member["WRAPPERS"]["d1"]["m1"] = [d1_m1_1_s2, d1_m1_1_s3, d1_m1_2_s2, d1_m1_2_s3,
                                                                  d1_m1_3_s2, d1_m1_3_s3, d1_m1_4_s2, d1_m1_4_s3]

        self.job_list._ordered_jobs_by_date_member["WRAPPERS"]["d1"]["m2"] = [d1_m2_1_s2, d1_m2_1_s3, d1_m2_2_s2, d1_m2_2_s3,
                                                                  d1_m2_3_s2, d1_m2_3_s3, d1_m2_4_s2, d1_m2_4_s3]

        section_list = [d1_m1_1_s2]

        self.job_packager.max_jobs = max_jobs
        self.job_packager._platform.max_wallclock = max_wallclock
        self.job_packager.wrapper_type = 'vertical'
        self.job_packager.jobs_in_wrapper = wrapper_expression
        self.job_packager.retrials = 0
        max_wrapper_job_by_section = {}
        max_wrapper_job_by_section["s1"] = max_wrapped_jobs
        max_wrapper_job_by_section["s2"] = max_wrapped_jobs
        max_wrapper_job_by_section["s3"] = max_wrapped_jobs
        max_wrapper_job_by_section["s4"] = max_wrapped_jobs
        wrapper_limits = dict()
        wrapper_limits["max"] = max_wrapped_jobs
        wrapper_limits["max_v"] = max_wrapped_jobs
        wrapper_limits["max_h"] = max_wrapped_jobs
        wrapper_limits["min"] = 2
        wrapper_limits["min_v"] = 2
        wrapper_limits["min_h"] = 2
        wrapper_limits["max_by_section"] = max_wrapper_job_by_section
        returned_packages = self.job_packager._build_vertical_packages(
            section_list, wrapper_limits)

        package_m1_s2_s3 = [d1_m1_1_s2, d1_m1_1_s3, d1_m1_2_s2, d1_m1_2_s3, d1_m1_3_s2, d1_m1_3_s3, d1_m1_4_s2,
                            d1_m1_4_s3]

        packages = [JobPackageVertical(package_m1_s2_s3,configuration=self.as_conf)]

        #returned_packages = returned_packages[0]
        for i in range(0, len(returned_packages)):
            self.assertListEqual(returned_packages[i]._jobs, packages[i]._jobs)

    def test_returned_packages_max_jobs_mixed_wrapper(self):
        wrapper_expression = "s2 s3"
        max_jobs = 10
        max_wrapped_jobs = 10
        max_wallclock = '10:00'

        date_list = ["d1"]
        member_list = ["m1", "m2"]
        chunk_list = [1, 2, 3, 4]
        for section,s_value in self.workflows['basic']['sections'].items():
            self.as_conf.jobs_data[section] = s_value
        self._createDummyJobs(
            self.workflows['basic'], date_list, member_list, chunk_list)

        self.job_list.get_job_by_name(
            'expid_d1_m1_s1').status = Status.COMPLETED
        self.job_list.get_job_by_name(
            'expid_d1_m2_s1').status = Status.COMPLETED

        self.job_list.get_job_by_name('expid_d1_m1_1_s2').status = Status.READY
        self.job_list.get_job_by_name('expid_d1_m2_1_s2').status = Status.READY

        d1_m1_1_s2 = self.job_list.get_job_by_name('expid_d1_m1_1_s2')
        d1_m1_2_s2 = self.job_list.get_job_by_name('expid_d1_m1_2_s2')
        d1_m1_3_s2 = self.job_list.get_job_by_name('expid_d1_m1_3_s2')
        d1_m1_4_s2 = self.job_list.get_job_by_name('expid_d1_m1_4_s2')
        d1_m2_1_s2 = self.job_list.get_job_by_name('expid_d1_m2_1_s2')
        d1_m2_2_s2 = self.job_list.get_job_by_name('expid_d1_m2_2_s2')
        d1_m2_3_s2 = self.job_list.get_job_by_name('expid_d1_m2_3_s2')
        d1_m2_4_s2 = self.job_list.get_job_by_name('expid_d1_m2_4_s2')

        d1_m1_1_s3 = self.job_list.get_job_by_name('expid_d1_m1_1_s3')
        d1_m1_2_s3 = self.job_list.get_job_by_name('expid_d1_m1_2_s3')
        d1_m1_3_s3 = self.job_list.get_job_by_name('expid_d1_m1_3_s3')
        d1_m1_4_s3 = self.job_list.get_job_by_name('expid_d1_m1_4_s3')
        d1_m2_1_s3 = self.job_list.get_job_by_name('expid_d1_m2_1_s3')
        d1_m2_2_s3 = self.job_list.get_job_by_name('expid_d1_m2_2_s3')
        d1_m2_3_s3 = self.job_list.get_job_by_name('expid_d1_m2_3_s3')
        d1_m2_4_s3 = self.job_list.get_job_by_name('expid_d1_m2_4_s3')

        self.job_list._ordered_jobs_by_date_member["WRAPPERS"]["d1"] = dict()
        self.job_list._ordered_jobs_by_date_member["WRAPPERS"]["d1"]["m1"] = [d1_m1_1_s2, d1_m1_1_s3, d1_m1_2_s2, d1_m1_2_s3,
                                                                  d1_m1_3_s2, d1_m1_3_s3, d1_m1_4_s2, d1_m1_4_s3]

        self.job_list._ordered_jobs_by_date_member["WRAPPERS"]["d1"]["m2"] = [d1_m2_1_s2, d1_m2_1_s3, d1_m2_2_s2, d1_m2_2_s3,
                                                                  d1_m2_3_s2, d1_m2_3_s3, d1_m2_4_s2, d1_m2_4_s3]

        section_list = [d1_m1_1_s2, d1_m2_1_s2]

        self.job_packager.max_jobs = max_jobs
        self.job_packager.retrials = 0
        self.job_packager._platform.max_wallclock = max_wallclock
        self.job_packager.wrapper_type = 'vertical'
        self.job_packager.jobs_in_wrapper = wrapper_expression
        max_wrapped_job_by_section = {}
        max_wrapped_job_by_section["s1"] = max_wrapped_jobs
        max_wrapped_job_by_section["s2"] = max_wrapped_jobs
        max_wrapped_job_by_section["s3"] = max_wrapped_jobs
        max_wrapped_job_by_section["s4"] = max_wrapped_jobs
        wrapper_limits = dict()
        wrapper_limits["max"] = max_wrapped_jobs
        wrapper_limits["max_v"] = max_wrapped_jobs
        wrapper_limits["max_h"] = max_wrapped_jobs
        wrapper_limits["min"] = 2
        wrapper_limits["min_v"] = 2
        wrapper_limits["min_h"] = 2
        wrapper_limits["max_by_section"] = max_wrapped_job_by_section
        returned_packages = self.job_packager._build_vertical_packages(
            section_list, wrapper_limits)

        package_m1_s2_s3 = [d1_m1_1_s2, d1_m1_1_s3, d1_m1_2_s2, d1_m1_2_s3, d1_m1_3_s2, d1_m1_3_s3, d1_m1_4_s2,
                            d1_m1_4_s3]
        package_m2_s2_s3 = [d1_m2_1_s2, d1_m2_1_s3, d1_m2_2_s2, d1_m2_2_s3, d1_m2_3_s2, d1_m2_3_s3, d1_m2_4_s2,
                            d1_m2_4_s3]

        packages = [JobPackageVertical(
            package_m1_s2_s3,configuration=self.as_conf), JobPackageVertical(package_m2_s2_s3,configuration=self.as_conf)]

        #returned_packages = returned_packages[0]
        # print("test_returned_packages_max_jobs_mixed_wrapper")
        for i in range(0, len(returned_packages)):
            # print("Element " + str(i))
            # print("Returned from packager")
            # for job in returned_packages[i]._jobs:
            #     print(job.name)
            # print("Build for test")
            # for _job in packages[i]._jobs:
            #     print(_job.name)
            self.assertListEqual(returned_packages[i]._jobs, packages[i]._jobs)

    def test_returned_packages_max_wrapped_jobs_mixed_wrapper(self):
        wrapper_expression = "s2 s3"
        max_jobs = 15
        max_wrapped_jobs = 5
        max_wallclock = '10:00'

        date_list = ["d1"]
        member_list = ["m1", "m2"]
        chunk_list = [1, 2, 3, 4]
        for section,s_value in self.workflows['basic']['sections'].items():
            self.as_conf.jobs_data[section] = s_value
        self._createDummyJobs(
            self.workflows['basic'], date_list, member_list, chunk_list)

        self.job_list.get_job_by_name(
            'expid_d1_m1_s1').status = Status.COMPLETED
        self.job_list.get_job_by_name(
            'expid_d1_m2_s1').status = Status.COMPLETED

        self.job_list.get_job_by_name('expid_d1_m1_1_s2').status = Status.READY
        self.job_list.get_job_by_name('expid_d1_m2_1_s2').status = Status.READY

        d1_m1_1_s2 = self.job_list.get_job_by_name('expid_d1_m1_1_s2')
        d1_m1_2_s2 = self.job_list.get_job_by_name('expid_d1_m1_2_s2')
        d1_m1_3_s2 = self.job_list.get_job_by_name('expid_d1_m1_3_s2')
        d1_m1_4_s2 = self.job_list.get_job_by_name('expid_d1_m1_4_s2')
        d1_m2_1_s2 = self.job_list.get_job_by_name('expid_d1_m2_1_s2')
        d1_m2_2_s2 = self.job_list.get_job_by_name('expid_d1_m2_2_s2')
        d1_m2_3_s2 = self.job_list.get_job_by_name('expid_d1_m2_3_s2')
        d1_m2_4_s2 = self.job_list.get_job_by_name('expid_d1_m2_4_s2')

        d1_m1_1_s3 = self.job_list.get_job_by_name('expid_d1_m1_1_s3')
        d1_m1_2_s3 = self.job_list.get_job_by_name('expid_d1_m1_2_s3')
        d1_m1_3_s3 = self.job_list.get_job_by_name('expid_d1_m1_3_s3')
        d1_m1_4_s3 = self.job_list.get_job_by_name('expid_d1_m1_4_s3')
        d1_m2_1_s3 = self.job_list.get_job_by_name('expid_d1_m2_1_s3')
        d1_m2_2_s3 = self.job_list.get_job_by_name('expid_d1_m2_2_s3')
        d1_m2_3_s3 = self.job_list.get_job_by_name('expid_d1_m2_3_s3')
        d1_m2_4_s3 = self.job_list.get_job_by_name('expid_d1_m2_4_s3')

        self.job_list._ordered_jobs_by_date_member["WRAPPERS"]["d1"] = dict()
        self.job_list._ordered_jobs_by_date_member["WRAPPERS"]["d1"]["m1"] = [d1_m1_1_s2, d1_m1_1_s3, d1_m1_2_s2, d1_m1_2_s3,
                                                                  d1_m1_3_s2, d1_m1_3_s3, d1_m1_4_s2, d1_m1_4_s3]

        self.job_list._ordered_jobs_by_date_member["WRAPPERS"]["d1"]["m2"] = [d1_m2_1_s2, d1_m2_1_s3, d1_m2_2_s2, d1_m2_2_s3,
                                                                  d1_m2_3_s2, d1_m2_3_s3, d1_m2_4_s2, d1_m2_4_s3]

        section_list = [d1_m1_1_s2, d1_m2_1_s2]

        self.job_packager.max_jobs = max_jobs
        self.job_packager.retrials = 0
        self.job_packager._platform.max_wallclock = max_wallclock
        self.job_packager.wrapper_type = 'vertical'
        self.job_packager.jobs_in_wrapper = wrapper_expression
        max_wrapped_job_by_section = {}
        max_wrapped_job_by_section["s1"] = max_wrapped_jobs
        max_wrapped_job_by_section["s2"] = max_wrapped_jobs
        max_wrapped_job_by_section["s3"] = max_wrapped_jobs
        max_wrapped_job_by_section["s4"] = max_wrapped_jobs
        wrapper_limits = dict()
        wrapper_limits["max"] = max_wrapped_jobs
        wrapper_limits["max_v"] = max_wrapped_jobs
        wrapper_limits["max_h"] = max_wrapped_jobs
        wrapper_limits["min"] = 2
        wrapper_limits["min_v"] = 2
        wrapper_limits["min_h"] = 2
        wrapper_limits["max_by_section"] = max_wrapped_job_by_section
        returned_packages = self.job_packager._build_vertical_packages(
            section_list, wrapper_limits)

        package_m1_s2_s3 = [d1_m1_1_s2, d1_m1_1_s3,
                            d1_m1_2_s2, d1_m1_2_s3, d1_m1_3_s2]
        package_m2_s2_s3 = [d1_m2_1_s2, d1_m2_1_s3,
                            d1_m2_2_s2, d1_m2_2_s3, d1_m2_3_s2]

        packages = [JobPackageVertical(
            package_m1_s2_s3,configuration=self.as_conf), JobPackageVertical(package_m2_s2_s3,configuration=self.as_conf)]

        #returned_packages = returned_packages[0]
        for i in range(0, len(returned_packages)):
            self.assertListEqual(returned_packages[i]._jobs, packages[i]._jobs)

    def test_returned_packages_max_wallclock_mixed_wrapper(self):
        date_list = ["d1"]
        member_list = ["m1", "m2"]
        chunk_list = [1, 2, 3, 4]
        for section,s_value in self.workflows['basic']['sections'].items():
            self.as_conf.jobs_data[section] = s_value
        self._createDummyJobs(
            self.workflows['basic'], date_list, member_list, chunk_list)

        self.job_list.get_job_by_name(
            'expid_d1_m1_s1').status = Status.COMPLETED
        self.job_list.get_job_by_name(
            'expid_d1_m2_s1').status = Status.COMPLETED

        self.job_list.get_job_by_name('expid_d1_m1_1_s2').status = Status.READY
        self.job_list.get_job_by_name('expid_d1_m2_1_s2').status = Status.READY

        wrapper_expression = "s2 s3"
        max_jobs = 18
        max_wrapped_jobs = 18
        max_wallclock = '01:00'

        d1_m1_1_s2 = self.job_list.get_job_by_name('expid_d1_m1_1_s2')
        d1_m1_2_s2 = self.job_list.get_job_by_name('expid_d1_m1_2_s2')
        d1_m1_3_s2 = self.job_list.get_job_by_name('expid_d1_m1_3_s2')
        d1_m1_4_s2 = self.job_list.get_job_by_name('expid_d1_m1_4_s2')
        d1_m2_1_s2 = self.job_list.get_job_by_name('expid_d1_m2_1_s2')
        d1_m2_2_s2 = self.job_list.get_job_by_name('expid_d1_m2_2_s2')
        d1_m2_3_s2 = self.job_list.get_job_by_name('expid_d1_m2_3_s2')
        d1_m2_4_s2 = self.job_list.get_job_by_name('expid_d1_m2_4_s2')

        d1_m1_1_s3 = self.job_list.get_job_by_name('expid_d1_m1_1_s3')
        d1_m1_2_s3 = self.job_list.get_job_by_name('expid_d1_m1_2_s3')
        d1_m1_3_s3 = self.job_list.get_job_by_name('expid_d1_m1_3_s3')
        d1_m1_4_s3 = self.job_list.get_job_by_name('expid_d1_m1_4_s3')
        d1_m2_1_s3 = self.job_list.get_job_by_name('expid_d1_m2_1_s3')
        d1_m2_2_s3 = self.job_list.get_job_by_name('expid_d1_m2_2_s3')
        d1_m2_3_s3 = self.job_list.get_job_by_name('expid_d1_m2_3_s3')
        d1_m2_4_s3 = self.job_list.get_job_by_name('expid_d1_m2_4_s3')

        self.job_list._ordered_jobs_by_date_member["WRAPPERS"]["d1"] = dict()
        self.job_list._ordered_jobs_by_date_member["WRAPPERS"]["d1"]["m1"] = [d1_m1_1_s2, d1_m1_1_s3, d1_m1_2_s2, d1_m1_2_s3,
                                                                  d1_m1_3_s2, d1_m1_3_s3, d1_m1_4_s2, d1_m1_4_s3]

        self.job_list._ordered_jobs_by_date_member["WRAPPERS"]["d1"]["m2"] = [d1_m2_1_s2, d1_m2_1_s3, d1_m2_2_s2, d1_m2_2_s3,
                                                                  d1_m2_3_s2, d1_m2_3_s3, d1_m2_4_s2, d1_m2_4_s3]

        section_list = [d1_m1_1_s2, d1_m2_1_s2]

        self.job_packager.max_jobs = max_jobs
        self.job_packager._platform.max_wallclock = max_wallclock
        self.job_packager.wrapper_type = 'vertical'
        self.job_packager.retrials = 0
        self.job_packager.jobs_in_wrapper = wrapper_expression
        max_wrapped_job_by_section = {}
        max_wrapped_job_by_section["s1"] = max_wrapped_jobs
        max_wrapped_job_by_section["s2"] = max_wrapped_jobs
        max_wrapped_job_by_section["s3"] = max_wrapped_jobs
        max_wrapped_job_by_section["s4"] = max_wrapped_jobs
        wrapper_limits = dict()
        wrapper_limits["max"] = max_wrapped_jobs
        wrapper_limits["max_v"] = max_wrapped_jobs
        wrapper_limits["max_h"] = max_wrapped_jobs
        wrapper_limits["min"] = 2
        wrapper_limits["min_v"] = 2
        wrapper_limits["min_h"] = 2
        wrapper_limits["max_by_section"] = max_wrapped_job_by_section
        returned_packages = self.job_packager._build_vertical_packages(
            section_list, wrapper_limits)

        package_m1_s2_s3 = [d1_m1_1_s2, d1_m1_1_s3, d1_m1_2_s2, d1_m1_2_s3]
        package_m2_s2_s3 = [d1_m2_1_s2, d1_m2_1_s3, d1_m2_2_s2, d1_m2_2_s3]

        packages = [JobPackageVertical(
            package_m1_s2_s3,configuration=self.as_conf), JobPackageVertical(package_m2_s2_s3,configuration=self.as_conf)]

        #returned_packages = returned_packages[0]
        for i in range(0, len(returned_packages)):
            self.assertListEqual(returned_packages[i]._jobs, packages[i]._jobs)

    def test_returned_packages_first_chunks_completed_mixed_wrapper(self):
        date_list = ["d1"]
        member_list = ["m1", "m2"]
        chunk_list = [1, 2, 3, 4]
        for section,s_value in self.workflows['basic']['sections'].items():
            self.as_conf.jobs_data[section] = s_value
        self._createDummyJobs(
            self.workflows['basic'], date_list, member_list, chunk_list)

        self.job_list.get_job_by_name(
            'expid_d1_m1_s1').status = Status.COMPLETED
        self.job_list.get_job_by_name(
            'expid_d1_m2_s1').status = Status.COMPLETED

        self.job_list.get_job_by_name(
            'expid_d1_m1_1_s2').status = Status.COMPLETED
        self.job_list.get_job_by_name(
            'expid_d1_m1_2_s2').status = Status.COMPLETED
        self.job_list.get_job_by_name(
            'expid_d1_m1_3_s2').status = Status.COMPLETED
        self.job_list.get_job_by_name(
            'expid_d1_m2_1_s2').status = Status.COMPLETED
        self.job_list.get_job_by_name(
            'expid_d1_m2_2_s2').status = Status.COMPLETED
        self.job_list.get_job_by_name(
            'expid_d1_m1_1_s3').status = Status.COMPLETED
        self.job_list.get_job_by_name(
            'expid_d1_m2_1_s3').status = Status.COMPLETED
        self.job_list.get_job_by_name(
            'expid_d1_m2_2_s3').status = Status.COMPLETED

        self.job_list.get_job_by_name('expid_d1_m1_4_s2').status = Status.READY
        self.job_list.get_job_by_name('expid_d1_m2_3_s2').status = Status.READY
        self.job_list.get_job_by_name('expid_d1_m1_2_s3').status = Status.READY

        d1_m1_1_s2 = self.job_list.get_job_by_name('expid_d1_m1_1_s2')
        d1_m1_2_s2 = self.job_list.get_job_by_name('expid_d1_m1_2_s2')
        d1_m1_3_s2 = self.job_list.get_job_by_name('expid_d1_m1_3_s2')
        d1_m1_4_s2 = self.job_list.get_job_by_name('expid_d1_m1_4_s2')
        d1_m2_1_s2 = self.job_list.get_job_by_name('expid_d1_m2_1_s2')
        d1_m2_2_s2 = self.job_list.get_job_by_name('expid_d1_m2_2_s2')
        d1_m2_3_s2 = self.job_list.get_job_by_name('expid_d1_m2_3_s2')
        d1_m2_4_s2 = self.job_list.get_job_by_name('expid_d1_m2_4_s2')

        d1_m1_1_s3 = self.job_list.get_job_by_name('expid_d1_m1_1_s3')
        d1_m1_2_s3 = self.job_list.get_job_by_name('expid_d1_m1_2_s3')
        d1_m1_3_s3 = self.job_list.get_job_by_name('expid_d1_m1_3_s3')
        d1_m1_4_s3 = self.job_list.get_job_by_name('expid_d1_m1_4_s3')
        d1_m2_1_s3 = self.job_list.get_job_by_name('expid_d1_m2_1_s3')
        d1_m2_2_s3 = self.job_list.get_job_by_name('expid_d1_m2_2_s3')
        d1_m2_3_s3 = self.job_list.get_job_by_name('expid_d1_m2_3_s3')
        d1_m2_4_s3 = self.job_list.get_job_by_name('expid_d1_m2_4_s3')

        self.job_list._ordered_jobs_by_date_member["WRAPPERS"]["d1"] = dict()
        self.job_list._ordered_jobs_by_date_member["WRAPPERS"]["d1"]["m1"] = [d1_m1_1_s2, d1_m1_1_s3, d1_m1_2_s2, d1_m1_2_s3, d1_m1_3_s2,
                                                                  d1_m1_3_s3, d1_m1_4_s2, d1_m1_4_s3]

        self.job_list._ordered_jobs_by_date_member["WRAPPERS"]["d1"]["m2"] = [d1_m2_1_s2, d1_m2_1_s3, d1_m2_2_s2, d1_m2_2_s3, d1_m2_3_s2,
                                                                  d1_m2_3_s3, d1_m2_4_s2, d1_m2_4_s3]

        wrapper_expression = "s2 s3"
        max_wrapped_jobs = 18
        max_jobs = 18
        max_wallclock = '10:00'

        section_list = [d1_m1_2_s3, d1_m1_4_s2, d1_m2_3_s2]

        self.job_packager.max_jobs = max_jobs
        self.job_packager._platform.max_wallclock = max_wallclock
        self.job_packager.wrapper_type = 'vertical'
        self.job_packager.retrials = 0
        self.job_packager.jobs_in_wrapper = wrapper_expression
        max_wrapped_job_by_section = {}
        max_wrapped_job_by_section["s1"] = max_wrapped_jobs
        max_wrapped_job_by_section["s2"] = max_wrapped_jobs
        max_wrapped_job_by_section["s3"] = max_wrapped_jobs
        max_wrapped_job_by_section["s4"] = max_wrapped_jobs
        wrapper_limits = dict()
        wrapper_limits["max"] = max_wrapped_jobs
        wrapper_limits["max_v"] = max_wrapped_jobs
        wrapper_limits["max_h"] = max_wrapped_jobs
        wrapper_limits["min"] = 2
        wrapper_limits["min_v"] = 2
        wrapper_limits["min_h"] = 2
        wrapper_limits["max_by_section"] = max_wrapped_job_by_section
        returned_packages = self.job_packager._build_vertical_packages(
            section_list, wrapper_limits)

        package_m1_s2_s3 = [d1_m1_2_s3, d1_m1_3_s3, d1_m1_4_s2, d1_m1_4_s3]
        package_m2_s2_s3 = [d1_m2_3_s2, d1_m2_3_s3, d1_m2_4_s2, d1_m2_4_s3]

        packages = [JobPackageVertical(
            package_m1_s2_s3,configuration=self.as_conf), JobPackageVertical(package_m2_s2_s3,configuration=self.as_conf)]

        #returned_packages = returned_packages[0]
        for i in range(0, len(returned_packages)):
            self.assertListEqual(returned_packages[i]._jobs, packages[i]._jobs)

    def test_ordered_dict_jobs_simple_workflow_mixed_wrapper(self):
        date_list = ["d1"]
        member_list = ["m1", "m2"]
        chunk_list = [1, 2, 3, 4]
        for section,s_value in self.workflows['basic']['sections'].items():
            self.as_conf.jobs_data[section] = s_value
        self._createDummyJobs(
            self.workflows['basic'], date_list, member_list, chunk_list)

        self.job_list.get_job_by_name(
            'expid_d1_m1_s1').status = Status.COMPLETED
        self.job_list.get_job_by_name(
            'expid_d1_m2_s1').status = Status.COMPLETED

        self.job_list.get_job_by_name('expid_d1_m1_1_s2').status = Status.READY
        self.job_list.get_job_by_name('expid_d1_m2_1_s2').status = Status.READY

        d1_m1_1_s2 = self.job_list.get_job_by_name('expid_d1_m1_1_s2')
        d1_m1_2_s2 = self.job_list.get_job_by_name('expid_d1_m1_2_s2')
        d1_m1_3_s2 = self.job_list.get_job_by_name('expid_d1_m1_3_s2')
        d1_m1_4_s2 = self.job_list.get_job_by_name('expid_d1_m1_4_s2')
        d1_m2_1_s2 = self.job_list.get_job_by_name('expid_d1_m2_1_s2')
        d1_m2_2_s2 = self.job_list.get_job_by_name('expid_d1_m2_2_s2')
        d1_m2_3_s2 = self.job_list.get_job_by_name('expid_d1_m2_3_s2')
        d1_m2_4_s2 = self.job_list.get_job_by_name('expid_d1_m2_4_s2')

        d1_m1_1_s3 = self.job_list.get_job_by_name('expid_d1_m1_1_s3')
        d1_m1_2_s3 = self.job_list.get_job_by_name('expid_d1_m1_2_s3')
        d1_m1_3_s3 = self.job_list.get_job_by_name('expid_d1_m1_3_s3')
        d1_m1_4_s3 = self.job_list.get_job_by_name('expid_d1_m1_4_s3')
        d1_m2_1_s3 = self.job_list.get_job_by_name('expid_d1_m2_1_s3')
        d1_m2_2_s3 = self.job_list.get_job_by_name('expid_d1_m2_2_s3')
        d1_m2_3_s3 = self.job_list.get_job_by_name('expid_d1_m2_3_s3')
        d1_m2_4_s3 = self.job_list.get_job_by_name('expid_d1_m2_4_s3')

        self.parser_mock.has_option = MagicMock(return_value=True)
        self.parser_mock.get = MagicMock(return_value="chunk")
        self.job_list._get_date = MagicMock(return_value='d1')

        ordered_jobs_by_date_member = dict()
        ordered_jobs_by_date_member["d1"] = dict()
        ordered_jobs_by_date_member["d1"]["m1"] = [d1_m1_1_s2, d1_m1_1_s3, d1_m1_2_s2, d1_m1_2_s3, d1_m1_3_s2,
                                                   d1_m1_3_s3, d1_m1_4_s2, d1_m1_4_s3]

        ordered_jobs_by_date_member["d1"]["m2"] = [d1_m2_1_s2, d1_m2_1_s3, d1_m2_2_s2, d1_m2_2_s3, d1_m2_3_s2,
                                                   d1_m2_3_s3, d1_m2_4_s2, d1_m2_4_s3]

        self.assertDictEqual(self.job_list._create_sorted_dict_jobs(
            "s2 s3"), ordered_jobs_by_date_member)

    def test_ordered_dict_jobs_running_date_mixed_wrapper(self):
        date_list = ["d1", "d2"]
        member_list = ["m1", "m2"]
        chunk_list = [1, 2, 3, 4]
        for section,s_value in self.workflows['running_date']['sections'].items():
            self.as_conf.jobs_data[section] = s_value
        self._createDummyJobs(
            self.workflows['running_date'], date_list, member_list, chunk_list)

        self.parser_mock.has_option = MagicMock(return_value=True)
        self.parser_mock.get = MagicMock(side_effect=["chunk", "chunk", "date"])
        self.job_list._get_date = MagicMock(side_effect=['d1', 'd2'])

        d1_m1_1_s2 = self.job_list.get_job_by_name('expid_d1_m1_1_s2')
        d1_m1_2_s2 = self.job_list.get_job_by_name('expid_d1_m1_2_s2')
        d1_m1_3_s2 = self.job_list.get_job_by_name('expid_d1_m1_3_s2')
        d1_m1_4_s2 = self.job_list.get_job_by_name('expid_d1_m1_4_s2')
        d1_m2_1_s2 = self.job_list.get_job_by_name('expid_d1_m2_1_s2')
        d1_m2_2_s2 = self.job_list.get_job_by_name('expid_d1_m2_2_s2')
        d1_m2_3_s2 = self.job_list.get_job_by_name('expid_d1_m2_3_s2')
        d1_m2_4_s2 = self.job_list.get_job_by_name('expid_d1_m2_4_s2')

        d1_m1_1_s3 = self.job_list.get_job_by_name('expid_d1_m1_1_s3')
        d1_m1_2_s3 = self.job_list.get_job_by_name('expid_d1_m1_2_s3')
        d1_m1_3_s3 = self.job_list.get_job_by_name('expid_d1_m1_3_s3')
        d1_m1_4_s3 = self.job_list.get_job_by_name('expid_d1_m1_4_s3')
        d1_m2_1_s3 = self.job_list.get_job_by_name('expid_d1_m2_1_s3')
        d1_m2_2_s3 = self.job_list.get_job_by_name('expid_d1_m2_2_s3')
        d1_m2_3_s3 = self.job_list.get_job_by_name('expid_d1_m2_3_s3')
        d1_m2_4_s3 = self.job_list.get_job_by_name('expid_d1_m2_4_s3')

        d1_s5 = self.job_list.get_job_by_name('expid_d1_s5')

        d2_m1_1_s2 = self.job_list.get_job_by_name('expid_d2_m1_1_s2')
        d2_m1_2_s2 = self.job_list.get_job_by_name('expid_d2_m1_2_s2')
        d2_m1_3_s2 = self.job_list.get_job_by_name('expid_d2_m1_3_s2')
        d2_m1_4_s2 = self.job_list.get_job_by_name('expid_d2_m1_4_s2')
        d2_m2_1_s2 = self.job_list.get_job_by_name('expid_d2_m2_1_s2')
        d2_m2_2_s2 = self.job_list.get_job_by_name('expid_d2_m2_2_s2')
        d2_m2_3_s2 = self.job_list.get_job_by_name('expid_d2_m2_3_s2')
        d2_m2_4_s2 = self.job_list.get_job_by_name('expid_d2_m2_4_s2')

        d2_m1_1_s3 = self.job_list.get_job_by_name('expid_d2_m1_1_s3')
        d2_m1_2_s3 = self.job_list.get_job_by_name('expid_d2_m1_2_s3')
        d2_m1_3_s3 = self.job_list.get_job_by_name('expid_d2_m1_3_s3')
        d2_m1_4_s3 = self.job_list.get_job_by_name('expid_d2_m1_4_s3')
        d2_m2_1_s3 = self.job_list.get_job_by_name('expid_d2_m2_1_s3')
        d2_m2_2_s3 = self.job_list.get_job_by_name('expid_d2_m2_2_s3')
        d2_m2_3_s3 = self.job_list.get_job_by_name('expid_d2_m2_3_s3')
        d2_m2_4_s3 = self.job_list.get_job_by_name('expid_d2_m2_4_s3')

        d2_s5 = self.job_list.get_job_by_name('expid_d2_s5')

        ordered_jobs_by_date_member = dict()
        ordered_jobs_by_date_member["d1"] = dict()
        ordered_jobs_by_date_member["d1"]["m1"] = [d1_m1_1_s2, d1_m1_1_s3, d1_m1_2_s2, d1_m1_2_s3, d1_m1_3_s2,
                                                   d1_m1_3_s3, d1_m1_4_s2, d1_m1_4_s3]

        ordered_jobs_by_date_member["d1"]["m2"] = [d1_m2_1_s2, d1_m2_1_s3, d1_m2_2_s2, d1_m2_2_s3, d1_m2_3_s2,
                                                   d1_m2_3_s3, d1_m2_4_s2, d1_m2_4_s3, d1_s5]
        ordered_jobs_by_date_member["d2"] = dict()
        ordered_jobs_by_date_member["d2"]["m1"] = [d2_m1_1_s2, d2_m1_1_s3, d2_m1_2_s2, d2_m1_2_s3, d2_m1_3_s2,
                                                   d2_m1_3_s3, d2_m1_4_s2, d2_m1_4_s3]

        ordered_jobs_by_date_member["d2"]["m2"] = [d2_m2_1_s2, d2_m2_1_s3, d2_m2_2_s2, d2_m2_2_s3, d2_m2_3_s2,
                                                   d2_m2_3_s3, d2_m2_4_s2, d2_m2_4_s3, d2_s5]

        self.assertDictEqual(self.job_list._create_sorted_dict_jobs(
            "s2 s3 s5"), ordered_jobs_by_date_member)

    def test_ordered_dict_jobs_running_once_mixed_wrapper(self):
        date_list = ["d1", "d2"]
        member_list = ["m1", "m2"]
        chunk_list = [1, 2, 3, 4]
        for section,s_value in self.workflows['running_once']['sections'].items():
            self.as_conf.jobs_data[section] = s_value
        self._createDummyJobs(
            self.workflows['running_once'], date_list, member_list, chunk_list)

        self.parser_mock.has_option = MagicMock(return_value=True)
        self.parser_mock.get = MagicMock(side_effect=["chunk", "chunk", "once"])
        self.job_list._get_date = MagicMock(side_effect=['d2', 'd1', 'd2'])

        d1_m1_1_s2 = self.job_list.get_job_by_name('expid_d1_m1_1_s2')
        d1_m1_2_s2 = self.job_list.get_job_by_name('expid_d1_m1_2_s2')
        d1_m1_3_s2 = self.job_list.get_job_by_name('expid_d1_m1_3_s2')
        d1_m1_4_s2 = self.job_list.get_job_by_name('expid_d1_m1_4_s2')
        d1_m2_1_s2 = self.job_list.get_job_by_name('expid_d1_m2_1_s2')
        d1_m2_2_s2 = self.job_list.get_job_by_name('expid_d1_m2_2_s2')
        d1_m2_3_s2 = self.job_list.get_job_by_name('expid_d1_m2_3_s2')
        d1_m2_4_s2 = self.job_list.get_job_by_name('expid_d1_m2_4_s2')

        d1_m1_1_s3 = self.job_list.get_job_by_name('expid_d1_m1_1_s3')
        d1_m1_2_s3 = self.job_list.get_job_by_name('expid_d1_m1_2_s3')
        d1_m1_3_s3 = self.job_list.get_job_by_name('expid_d1_m1_3_s3')
        d1_m1_4_s3 = self.job_list.get_job_by_name('expid_d1_m1_4_s3')
        d1_m2_1_s3 = self.job_list.get_job_by_name('expid_d1_m2_1_s3')
        d1_m2_2_s3 = self.job_list.get_job_by_name('expid_d1_m2_2_s3')
        d1_m2_3_s3 = self.job_list.get_job_by_name('expid_d1_m2_3_s3')
        d1_m2_4_s3 = self.job_list.get_job_by_name('expid_d1_m2_4_s3')

        d2_m1_1_s2 = self.job_list.get_job_by_name('expid_d2_m1_1_s2')
        d2_m1_2_s2 = self.job_list.get_job_by_name('expid_d2_m1_2_s2')
        d2_m1_3_s2 = self.job_list.get_job_by_name('expid_d2_m1_3_s2')
        d2_m1_4_s2 = self.job_list.get_job_by_name('expid_d2_m1_4_s2')
        d2_m2_1_s2 = self.job_list.get_job_by_name('expid_d2_m2_1_s2')
        d2_m2_2_s2 = self.job_list.get_job_by_name('expid_d2_m2_2_s2')
        d2_m2_3_s2 = self.job_list.get_job_by_name('expid_d2_m2_3_s2')
        d2_m2_4_s2 = self.job_list.get_job_by_name('expid_d2_m2_4_s2')

        d2_m1_1_s3 = self.job_list.get_job_by_name('expid_d2_m1_1_s3')
        d2_m1_2_s3 = self.job_list.get_job_by_name('expid_d2_m1_2_s3')
        d2_m1_3_s3 = self.job_list.get_job_by_name('expid_d2_m1_3_s3')
        d2_m1_4_s3 = self.job_list.get_job_by_name('expid_d2_m1_4_s3')
        d2_m2_1_s3 = self.job_list.get_job_by_name('expid_d2_m2_1_s3')
        d2_m2_2_s3 = self.job_list.get_job_by_name('expid_d2_m2_2_s3')
        d2_m2_3_s3 = self.job_list.get_job_by_name('expid_d2_m2_3_s3')
        d2_m2_4_s3 = self.job_list.get_job_by_name('expid_d2_m2_4_s3')

        s5 = self.job_list.get_job_by_name('expid_s5')

        ordered_jobs_by_date_member = dict()
        ordered_jobs_by_date_member["d1"] = dict()
        ordered_jobs_by_date_member["d1"]["m1"] = [d1_m1_1_s2, d1_m1_1_s3, d1_m1_2_s2, d1_m1_2_s3, d1_m1_3_s2,
                                                   d1_m1_3_s3, d1_m1_4_s2, d1_m1_4_s3]

        ordered_jobs_by_date_member["d1"]["m2"] = [d1_m2_1_s2, d1_m2_1_s3, d1_m2_2_s2, d1_m2_2_s3, d1_m2_3_s2,
                                                   d1_m2_3_s3, d1_m2_4_s2, d1_m2_4_s3]
        ordered_jobs_by_date_member["d2"] = dict()
        ordered_jobs_by_date_member["d2"]["m1"] = [d2_m1_1_s2, d2_m1_1_s3, d2_m1_2_s2, d2_m1_2_s3, d2_m1_3_s2,
                                                   d2_m1_3_s3, d2_m1_4_s2, d2_m1_4_s3]

        ordered_jobs_by_date_member["d2"]["m2"] = [d2_m2_1_s2, d2_m2_1_s3, d2_m2_2_s2, d2_m2_2_s3, d2_m2_3_s2,
                                                   d2_m2_3_s3, d2_m2_4_s2, d2_m2_4_s3, s5]

        self.assertDictEqual(self.job_list._create_sorted_dict_jobs(
            "s2 s3 s5"), ordered_jobs_by_date_member)

    def test_ordered_dict_jobs_synchronize_date_mixed_wrapper(self):
        date_list = ["d1", "d2"]
        member_list = ["m1", "m2"]
        chunk_list = [1, 2, 3, 4]
        for section,s_value in self.workflows['synchronize_date']['sections'].items():
            self.as_conf.jobs_data[section] = s_value
        self._createDummyJobs(
            self.workflows['synchronize_date'], date_list, member_list, chunk_list)

        self.parser_mock.has_option = MagicMock(return_value=True)
        self.parser_mock.get = MagicMock(return_value="chunk")
        self.job_list._get_date = MagicMock(
            side_effect=['d2', 'd2', 'd2', 'd2', 'd1', 'd2'])

        d1_m1_1_s2 = self.job_list.get_job_by_name('expid_d1_m1_1_s2')
        d1_m1_2_s2 = self.job_list.get_job_by_name('expid_d1_m1_2_s2')
        d1_m1_3_s2 = self.job_list.get_job_by_name('expid_d1_m1_3_s2')
        d1_m1_4_s2 = self.job_list.get_job_by_name('expid_d1_m1_4_s2')
        d1_m2_1_s2 = self.job_list.get_job_by_name('expid_d1_m2_1_s2')
        d1_m2_2_s2 = self.job_list.get_job_by_name('expid_d1_m2_2_s2')
        d1_m2_3_s2 = self.job_list.get_job_by_name('expid_d1_m2_3_s2')
        d1_m2_4_s2 = self.job_list.get_job_by_name('expid_d1_m2_4_s2')

        d1_m1_1_s3 = self.job_list.get_job_by_name('expid_d1_m1_1_s3')
        d1_m1_2_s3 = self.job_list.get_job_by_name('expid_d1_m1_2_s3')
        d1_m1_3_s3 = self.job_list.get_job_by_name('expid_d1_m1_3_s3')
        d1_m1_4_s3 = self.job_list.get_job_by_name('expid_d1_m1_4_s3')
        d1_m2_1_s3 = self.job_list.get_job_by_name('expid_d1_m2_1_s3')
        d1_m2_2_s3 = self.job_list.get_job_by_name('expid_d1_m2_2_s3')
        d1_m2_3_s3 = self.job_list.get_job_by_name('expid_d1_m2_3_s3')
        d1_m2_4_s3 = self.job_list.get_job_by_name('expid_d1_m2_4_s3')

        d2_m1_1_s2 = self.job_list.get_job_by_name('expid_d2_m1_1_s2')
        d2_m1_2_s2 = self.job_list.get_job_by_name('expid_d2_m1_2_s2')
        d2_m1_3_s2 = self.job_list.get_job_by_name('expid_d2_m1_3_s2')
        d2_m1_4_s2 = self.job_list.get_job_by_name('expid_d2_m1_4_s2')
        d2_m2_1_s2 = self.job_list.get_job_by_name('expid_d2_m2_1_s2')
        d2_m2_2_s2 = self.job_list.get_job_by_name('expid_d2_m2_2_s2')
        d2_m2_3_s2 = self.job_list.get_job_by_name('expid_d2_m2_3_s2')
        d2_m2_4_s2 = self.job_list.get_job_by_name('expid_d2_m2_4_s2')

        d2_m1_1_s3 = self.job_list.get_job_by_name('expid_d2_m1_1_s3')
        d2_m1_2_s3 = self.job_list.get_job_by_name('expid_d2_m1_2_s3')
        d2_m1_3_s3 = self.job_list.get_job_by_name('expid_d2_m1_3_s3')
        d2_m1_4_s3 = self.job_list.get_job_by_name('expid_d2_m1_4_s3')
        d2_m2_1_s3 = self.job_list.get_job_by_name('expid_d2_m2_1_s3')
        d2_m2_2_s3 = self.job_list.get_job_by_name('expid_d2_m2_2_s3')
        d2_m2_3_s3 = self.job_list.get_job_by_name('expid_d2_m2_3_s3')
        d2_m2_4_s3 = self.job_list.get_job_by_name('expid_d2_m2_4_s3')

        _1_s5 = self.job_list.get_job_by_name('expid_1_s5')
        _2_s5 = self.job_list.get_job_by_name('expid_2_s5')
        _3_s5 = self.job_list.get_job_by_name('expid_3_s5')
        _4_s5 = self.job_list.get_job_by_name('expid_4_s5')

        ordered_jobs_by_date_member = dict()
        ordered_jobs_by_date_member["d1"] = dict()
        ordered_jobs_by_date_member["d1"]["m1"] = [d1_m1_1_s2, d1_m1_1_s3, d1_m1_2_s2, d1_m1_2_s3, d1_m1_3_s2,
                                                   d1_m1_3_s3, d1_m1_4_s2, d1_m1_4_s3]

        ordered_jobs_by_date_member["d1"]["m2"] = [d1_m2_1_s2, d1_m2_1_s3, d1_m2_2_s2, d1_m2_2_s3, d1_m2_3_s2,
                                                   d1_m2_3_s3, d1_m2_4_s2, d1_m2_4_s3]
        ordered_jobs_by_date_member["d2"] = dict()
        ordered_jobs_by_date_member["d2"]["m1"] = [d2_m1_1_s2, d2_m1_1_s3, d2_m1_2_s2, d2_m1_2_s3, d2_m1_3_s2,
                                                   d2_m1_3_s3, d2_m1_4_s2, d2_m1_4_s3]

        ordered_jobs_by_date_member["d2"]["m2"] = [d2_m2_1_s2, d2_m2_1_s3, _1_s5, d2_m2_2_s2, d2_m2_2_s3, _2_s5, d2_m2_3_s2,
                                                   d2_m2_3_s3, _3_s5, d2_m2_4_s2, d2_m2_4_s3, _4_s5]

        self.assertDictEqual(self.job_list._create_sorted_dict_jobs(
            "s2 s3 s5"), ordered_jobs_by_date_member)

    def test_ordered_dict_jobs_synchronize_member_mixed_wrapper(self):
        date_list = ["d1", "d2"]
        member_list = ["m1", "m2"]
        chunk_list = [1, 2, 3, 4]
        for section,s_value in self.workflows['synchronize_member']['sections'].items():
            self.as_conf.jobs_data[section] = s_value
        self._createDummyJobs(
            self.workflows['synchronize_member'], date_list, member_list, chunk_list)

        self.parser_mock.has_option = MagicMock(return_value=True)
        self.parser_mock.get = MagicMock(return_value="chunk")
        self.job_list._get_date = MagicMock(side_effect=['d1', 'd2'])

        d1_m1_1_s2 = self.job_list.get_job_by_name('expid_d1_m1_1_s2')
        d1_m1_2_s2 = self.job_list.get_job_by_name('expid_d1_m1_2_s2')
        d1_m1_3_s2 = self.job_list.get_job_by_name('expid_d1_m1_3_s2')
        d1_m1_4_s2 = self.job_list.get_job_by_name('expid_d1_m1_4_s2')
        d1_m2_1_s2 = self.job_list.get_job_by_name('expid_d1_m2_1_s2')
        d1_m2_2_s2 = self.job_list.get_job_by_name('expid_d1_m2_2_s2')
        d1_m2_3_s2 = self.job_list.get_job_by_name('expid_d1_m2_3_s2')
        d1_m2_4_s2 = self.job_list.get_job_by_name('expid_d1_m2_4_s2')

        d1_m1_1_s3 = self.job_list.get_job_by_name('expid_d1_m1_1_s3')
        d1_m1_2_s3 = self.job_list.get_job_by_name('expid_d1_m1_2_s3')
        d1_m1_3_s3 = self.job_list.get_job_by_name('expid_d1_m1_3_s3')
        d1_m1_4_s3 = self.job_list.get_job_by_name('expid_d1_m1_4_s3')
        d1_m2_1_s3 = self.job_list.get_job_by_name('expid_d1_m2_1_s3')
        d1_m2_2_s3 = self.job_list.get_job_by_name('expid_d1_m2_2_s3')
        d1_m2_3_s3 = self.job_list.get_job_by_name('expid_d1_m2_3_s3')
        d1_m2_4_s3 = self.job_list.get_job_by_name('expid_d1_m2_4_s3')

        d2_m1_1_s2 = self.job_list.get_job_by_name('expid_d2_m1_1_s2')
        d2_m1_2_s2 = self.job_list.get_job_by_name('expid_d2_m1_2_s2')
        d2_m1_3_s2 = self.job_list.get_job_by_name('expid_d2_m1_3_s2')
        d2_m1_4_s2 = self.job_list.get_job_by_name('expid_d2_m1_4_s2')
        d2_m2_1_s2 = self.job_list.get_job_by_name('expid_d2_m2_1_s2')
        d2_m2_2_s2 = self.job_list.get_job_by_name('expid_d2_m2_2_s2')
        d2_m2_3_s2 = self.job_list.get_job_by_name('expid_d2_m2_3_s2')
        d2_m2_4_s2 = self.job_list.get_job_by_name('expid_d2_m2_4_s2')

        d2_m1_1_s3 = self.job_list.get_job_by_name('expid_d2_m1_1_s3')
        d2_m1_2_s3 = self.job_list.get_job_by_name('expid_d2_m1_2_s3')
        d2_m1_3_s3 = self.job_list.get_job_by_name('expid_d2_m1_3_s3')
        d2_m1_4_s3 = self.job_list.get_job_by_name('expid_d2_m1_4_s3')
        d2_m2_1_s3 = self.job_list.get_job_by_name('expid_d2_m2_1_s3')
        d2_m2_2_s3 = self.job_list.get_job_by_name('expid_d2_m2_2_s3')
        d2_m2_3_s3 = self.job_list.get_job_by_name('expid_d2_m2_3_s3')
        d2_m2_4_s3 = self.job_list.get_job_by_name('expid_d2_m2_4_s3')

        d1_1_s5 = self.job_list.get_job_by_name('expid_d1_1_s5')
        d1_2_s5 = self.job_list.get_job_by_name('expid_d1_2_s5')
        d1_3_s5 = self.job_list.get_job_by_name('expid_d1_3_s5')
        d1_4_s5 = self.job_list.get_job_by_name('expid_d1_4_s5')

        d2_1_s5 = self.job_list.get_job_by_name('expid_d2_1_s5')
        d2_2_s5 = self.job_list.get_job_by_name('expid_d2_2_s5')
        d2_3_s5 = self.job_list.get_job_by_name('expid_d2_3_s5')
        d2_4_s5 = self.job_list.get_job_by_name('expid_d2_4_s5')

        ordered_jobs_by_date_member = dict()
        ordered_jobs_by_date_member["d1"] = dict()
        ordered_jobs_by_date_member["d1"]["m1"] = [d1_m1_1_s2, d1_m1_1_s3, d1_m1_2_s2, d1_m1_2_s3, d1_m1_3_s2,
                                                   d1_m1_3_s3, d1_m1_4_s2, d1_m1_4_s3]

        ordered_jobs_by_date_member["d1"]["m2"] = [d1_m2_1_s2, d1_m2_1_s3, d1_1_s5, d1_m2_2_s2, d1_m2_2_s3, d1_2_s5, d1_m2_3_s2,
                                                   d1_m2_3_s3, d1_3_s5, d1_m2_4_s2, d1_m2_4_s3, d1_4_s5]
        ordered_jobs_by_date_member["d2"] = dict()
        ordered_jobs_by_date_member["d2"]["m1"] = [d2_m1_1_s2, d2_m1_1_s3, d2_m1_2_s2, d2_m1_2_s3, d2_m1_3_s2,
                                                   d2_m1_3_s3, d2_m1_4_s2, d2_m1_4_s3]

        ordered_jobs_by_date_member["d2"]["m2"] = [d2_m2_1_s2, d2_m2_1_s3, d2_1_s5, d2_m2_2_s2, d2_m2_2_s3, d2_2_s5, d2_m2_3_s2,
                                                   d2_m2_3_s3, d2_3_s5, d2_m2_4_s2, d2_m2_4_s3, d2_4_s5]

        self.assertDictEqual(self.job_list._create_sorted_dict_jobs(
            "s2 s3 s5"), ordered_jobs_by_date_member)

    def _createDummyJobs(self, sections_dict, date_list, member_list, chunk_list):
        for section, section_dict in sections_dict.get('sections').items():
            running = section_dict['RUNNING']
            wallclock = section_dict['WALLCLOCK']

            if running == 'once':
                name = 'expid_' + section
                job = self._createDummyJob(name, wallclock, section)
                self.job_list._job_list.append(job)
            elif running == 'date':
                for date in date_list:
                    name = 'expid_' + date + "_" + section
                    job = self._createDummyJob(name, wallclock, section, date)
                    self.job_list._job_list.append(job)
            elif running == 'member':
                for date in date_list:
                    for member in member_list:
                        name = 'expid_' + date + "_" + member + "_" + section
                        job = self._createDummyJob(
                            name, wallclock, section, date, member)
                        self.job_list._job_list.append(job)
            elif running == 'chunk':
                synchronize_type = section_dict['SYNCHRONIZE'] if 'SYNCHRONIZE' in section_dict else None
                if synchronize_type == 'date':
                    for chunk in chunk_list:
                        name = 'expid_' + str(chunk) + "_" + section
                        job = self._createDummyJob(
                            name, wallclock, section, None, None, chunk)
                        self.job_list._job_list.append(job)
                elif synchronize_type == 'member':
                    for date in date_list:
                        for chunk in chunk_list:
                            name = 'expid_' + date + "_" + \
                                str(chunk) + "_" + section
                            job = self._createDummyJob(
                                name, wallclock, section, date, None, chunk)
                            self.job_list._job_list.append(job)
                else:
                    for date in date_list:
                        for member in member_list:
                            for chunk in chunk_list:
                                name = 'expid_' + date + "_" + member + \
                                    "_" + str(chunk) + "_" + section
                                job = self._createDummyJob(
                                    name, wallclock, section, date, member, chunk)
                                self.job_list._job_list.append(job)

        self.job_list._date_list = date_list
        self.job_list._member_list = member_list
        self.job_list._chunk_list = chunk_list

        self.job_list._dic_jobs = DicJobs(
            self.job_list, date_list, member_list, chunk_list, "", 0,jobs_data={},experiment_data=self.as_conf.experiment_data)
        self._manage_dependencies(sections_dict)

    def _manage_dependencies(self, sections_dict):
        for job in self.job_list.get_job_list():
            section = job.section
            dependencies = sections_dict['sections'][section][
                'DEPENDENCIES'] if 'DEPENDENCIES' in sections_dict['sections'][section] else ''
            self._manage_job_dependencies(job, dependencies, sections_dict)

    def _manage_job_dependencies(self, job, dependencies, sections_dict):
        for key in dependencies.split():
            if '-' not in key:
                dependency = Dependency(key)
            else:
                sign = '-' if '-' in key else '+'
                key_split = key.split(sign)
                section = key_split[0]
                distance = key_split[1]
                dependency_running_type = sections_dict['sections'][section]['RUNNING']
                dependency = Dependency(section, int(
                    distance), dependency_running_type, sign)

            skip, (chunk, member, date) = self.job_list._calculate_dependency_metadata(job.chunk, self.job_list.get_chunk_list(),
                                                                                       job.member, self.job_list.get_member_list(),
                                                                                       job.date, self.job_list.get_date_list(),
                                                                                       dependency)
            if skip:
                continue

            for parent in self._filter_jobs(dependency.section, date, member, chunk):
                job.add_parent(parent)

    def _filter_jobs(self, section, date=None, member=None, chunk=None):
        # TODO: improve the efficiency
        jobs = [job for job in self.job_list.get_job_list() if job.section == section and job.date == date and job.member == member and job.chunk == chunk]
        return jobs

    def _createDummyJob(self, name, total_wallclock, section, date=None, member=None, chunk=None):
        job_id = randrange(1, 999)
        job = Job(name, job_id, Status.WAITING, 0)
        job.type = randrange(0, 2)
        job.packed = False
        job.hold = False
        job.wallclock = total_wallclock
        job.platform = self._platform

        job.date = date
        job.member = member
        job.chunk = chunk
        job.section = section

        return job

import inspect
class FakeBasicConfig:
    def __init__(self):
        pass
    def props(self):
        pr = {}
        for name in dir(self):
            value = getattr(self, name)
            if not name.startswith('__') and not inspect.ismethod(value) and not inspect.isfunction(value):
                pr[name] = value
        return pr
    DB_DIR = '/dummy/db/dir'
    DB_FILE = '/dummy/db/file'
    DB_PATH = '/dummy/db/path'
    LOCAL_ROOT_DIR = '/dummy/local/root/dir'
    LOCAL_TMP_DIR = '/dummy/local/temp/dir'
    LOCAL_PROJ_DIR = '/dummy/local/proj/dir'
    DEFAULT_PLATFORMS_CONF = ''
    DEFAULT_JOBS_CONF = ''



import unittest
from autosubmit.statistics.statistics import Statistics
from autosubmit.job.job_common import Status
from autosubmit.job.job_utils import SubJobManager, SubJob
from autosubmitconfigparser.config.basicconfig import BasicConfig
from autosubmitconfigparser.config.configcommon import AutosubmitConfig
from bscearth.utils.config_parser import ConfigParserFactory
from autosubmit.autosubmit import Autosubmit
from autosubmit.job.job_list import JobList
# import autosubmit.experiment.common_db_requests as DbRequests
import autosubmit.database.db_structure as DbStructure
# from autosubmit.database.db_jobdata import JobDataStructure, ExperimentGraphDrawing

@unittest.skip("TODO: looks like this test was used by devs to run an existing experiment a49z")
class TestStatistics(unittest.TestCase):
  def setUp(self):
    self.expid = "a49z"

  def test_normal_execution(self):
    print("Testing normal execution")
    expid = self.expid
    period_fi = ""
    period_ini = ""
    ft = "Any"
    results = None
    subjobs = list()
    BasicConfig.read()
    path_structure = BasicConfig.STRUCTURES_DIR
    path_local_root = BasicConfig.LOCAL_ROOT_DIR
    as_conf = AutosubmitConfig(expid)
    as_conf.reload(force_load=True)
    job_list = Autosubmit.load_job_list(expid, as_conf, False)
    jobs_considered = [job for job in job_list.get_job_list() if job.status not in [
            Status.READY, Status.WAITING]]
    job_to_package, package_to_jobs, _, _ = JobList.retrieve_packages(
            BasicConfig, expid, [job.name for job in job_list.get_job_list()])
    queue_time_fixes = {}
    if job_to_package:
      current_table_structure = DbStructure.get_structure(expid, BasicConfig.STRUCTURES_DIR)
      subjobs = []
      for job in job_list.get_job_list():
          job_info = JobList.retrieve_times(job.status, job.name, job._tmp_path, make_exception=False, job_times=None, seconds=True, job_data_collection=None)
          time_total = (job_info.queue_time + job_info.run_time) if job_info else 0
          subjobs.append(
              SubJob(job.name,
                  job_to_package.get(job.name, None),
                  job_info.queue_time if job_info else 0,
                  job_info.run_time if job_info else 0,
                  time_total,
                  job_info.status if job_info else Status.UNKNOWN)
          )
      queue_time_fixes = SubJobManager(subjobs, job_to_package, package_to_jobs, current_table_structure).get_collection_of_fixes_applied()


    if len(jobs_considered) > 0:
      print("Get results")
      exp_stats = Statistics(jobs_considered, period_ini, period_fi, queue_time_fixes)
      exp_stats.calculate_statistics()
      exp_stats.calculate_summary()
      exp_stats.make_old_format()    
      print(exp_stats.get_summary_as_list())
      failed_jobs_dict = exp_stats.build_failed_jobs_only_list() 
    else:
        raise Exception("Autosubmit API couldn't find jobs that match your search criteria (Section: {0}) in the period from {1} to {2}.".format(
            ft, period_ini, period_fi))
    return results

if __name__ == '__main__':
  unittest.main()


#!/usr/bin/python

# Copyright 2015-2020 Earth Sciences Department, BSC-CNS
# This file is part of Autosubmit.

# Autosubmit is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Autosubmit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Autosubmit.  If not, see <http://www.gnu.org/licenses/>.

import unittest
from collections import namedtuple
from autosubmit.history.data_classes.job_data import JobData
from autosubmit.history.strategies import PlatformInformationHandler, TwoDimWrapperDistributionStrategy
from autosubmit.history.platform_monitor.slurm_monitor import SlurmMonitor
job_dc = namedtuple("Job", ["job_name", "date", "member", "status_str", "children", "children_list"])

class Test2DWrapperDistributionStrategy(unittest.TestCase):
  def setUp(self):
    self.strategy = TwoDimWrapperDistributionStrategy()
    self.job_data_dcs_in_wrapper = [
      JobData(0, job_name="a29z_20000101_fc2_1_POSTR", status="COMPLETED", submit=10, start=100, finish=200, ncpus=100, energy=0, children="a29z_20000101_fc1_1_CLEAN, a29z_20000101_fc3_1_POST"),
      JobData(0, job_name="a29z_20000101_fc1_1_CLEAN", status="COMPLETED", submit=10, start=100, finish=200, ncpus=100, energy=0, children="a29z_20000101_fc2_1_CLEAN"),
      JobData(0, job_name="a29z_20000101_fc3_1_POST", status="COMPLETED", submit=10, start=100, finish=200, ncpus=100, energy=0, children="a29z_20000101_fc0_3_SIM"),
      JobData(0, job_name="a29z_20000101_fc2_1_CLEAN", status="COMPLETED", submit=10, start=100, finish=200, ncpus=100, energy=0, children=""),
      JobData(0, job_name="a29z_20000101_fc0_3_SIM", status="COMPLETED", submit=10, start=100, finish=200, ncpus=100, energy=0, children=""),
      JobData(0, job_name="a29z_20000101_fc1_2_POSTR1", status="COMPLETED", submit=10, start=100, finish=200, ncpus=100, energy=0, children="a29z_20000101_fc1_5_POST2"),
      JobData(0, job_name="a29z_20000101_fc1_5_POST2", status="COMPLETED", submit=10, start=100, finish=200, ncpus=100, energy=0, children="a29z_20000101_fc1_4_POST3"),
      JobData(0, job_name="a29z_20000101_fc1_4_POST3", status="COMPLETED", submit=10, start=100, finish=200, ncpus=100, energy=0, children="a29z_20000101_fc2_5_CLEAN4"),
      JobData(0, job_name="a29z_20000101_fc2_5_CLEAN4", status="COMPLETED", submit=10, start=100, finish=200, ncpus=100, energy=0, children="a29z_20000101_fc0_1_POST5"),
      JobData(0, job_name="a29z_20000101_fc0_1_POST5", status="COMPLETED", submit=10, start=100, finish=200, ncpus=100, energy=0, children=""),
    ]

  def test_get_all_children(self):    
    children = self.strategy._get_all_children(self.job_data_dcs_in_wrapper)  
    self.assertTrue(len(children) == 8)
  
  def test_get_roots(self):
    roots = self.strategy._get_roots(self.job_data_dcs_in_wrapper)    
    self.assertTrue(len(roots) == 2)

  def test_get_level(self):
    roots = self.strategy._get_roots(self.job_data_dcs_in_wrapper) 
    job_name_to_children_names = {job.job_name: job.children_list for job in self.job_data_dcs_in_wrapper}      
    next_level = self.strategy.get_level(roots, job_name_to_children_names)                
    self.assertTrue(len(next_level) == 3)

  def test_get_jobs_per_level(self):
    levels = self.strategy.get_jobs_per_level(self.job_data_dcs_in_wrapper)    
    for level in levels:
      print([job.job_name for job in level])
    self.assertTrue(len(levels) == 5)
    self.assertTrue("a29z_20000101_fc0_1_POST5" in [job.job_name for job in levels[4]])

  def test_energy_distribution(self):
    ssh_output = '''                 17857525  COMPLETED         10        1 2021-10-13T15:51:16 2021-10-13T15:51:17 2021-10-13T15:52:47         2.62K                                                     
           17857525.batch  COMPLETED         10        1 2021-10-13T15:51:17 2021-10-13T15:51:17 2021-10-13T15:52:47          1.88K                     6264K                     6264K 
          17857525.extern  COMPLETED         10        1 2021-10-13T15:51:17 2021-10-13T15:51:17 2021-10-13T15:52:47          1.66K                      473K                       68K 
               17857525.0  COMPLETED         10        1 2021-10-13T15:51:21 2021-10-13T15:51:21 2021-10-13T15:51:22            186                      352K                   312.30K 
               17857525.1  COMPLETED         10        1 2021-10-13T15:51:23 2021-10-13T15:51:23 2021-10-13T15:51:24            186                      420K                   306.70K 
               17857525.2  COMPLETED         10        1 2021-10-13T15:51:24 2021-10-13T15:51:24 2021-10-13T15:51:27            188                      352K                   325.80K 
               17857525.3  COMPLETED         10        1 2021-10-13T15:51:28 2021-10-13T15:51:28 2021-10-13T15:51:29            192                      352K                   341.90K
               17857525.4  COMPLETED         10        1 2021-10-13T15:51:28 2021-10-13T15:51:28 2021-10-13T15:51:29            210                      352K                   341.90K                
    '''
    slurm_monitor = SlurmMonitor(ssh_output)
    info_handler = PlatformInformationHandler(TwoDimWrapperDistributionStrategy())
    job_dcs = info_handler.execute_distribution(self.job_data_dcs_in_wrapper[0], self.job_data_dcs_in_wrapper, slurm_monitor)
    for job in job_dcs:
      print(("{0} -> {1} and {2} : ncpus {3} running {4}".format(job.job_name, job.energy, job.rowstatus, job.ncpus, job.running_time)))
    for level in info_handler.strategy.jobs_per_level:
      print([job.job_name for job in level])
    total_in_jobs = sum(job.energy for job in job_dcs[:-1]) # ignore last
    self.assertTrue(abs(total_in_jobs - slurm_monitor.total_energy) <= 10)
    self.assertTrue(abs(job_dcs[0].energy - 259) < 1)
    self.assertTrue(abs(job_dcs[1].energy - 259) < 1)
    self.assertTrue(abs(job_dcs[2].energy - 228) < 1)
    self.assertTrue(abs(job_dcs[3].energy - 228) < 1)
    self.assertTrue(abs(job_dcs[4].energy - 228) < 1)
    self.assertTrue(abs(job_dcs[5].energy - 228.67) < 1)
    self.assertTrue(abs(job_dcs[6].energy - 228.67) < 1)
    self.assertTrue(abs(job_dcs[7].energy - 228.67) < 1)
    self.assertTrue(abs(job_dcs[8].energy - 358) < 1)
    self.assertTrue(abs(job_dcs[9].energy - 376) < 1)

if __name__ == '__main__':
  unittest.main()


import copy
import inspect
import mock
import tempfile
import unittest
from copy import deepcopy
from datetime import datetime

from autosubmit.job.job import Job
from autosubmit.job.job_common import Status
from autosubmit.job.job_list import JobList
from autosubmit.job.job_list_persistence import JobListPersistenceDb
from autosubmitconfigparser.config.yamlparser import YAMLParserFactory


class FakeBasicConfig:
    def __init__(self):
        pass
    def props(self):
        pr = {}
        for name in dir(self):
            value = getattr(self, name)
            if not name.startswith('__') and not inspect.ismethod(value) and not inspect.isfunction(value):
                pr[name] = value
        return pr
    DB_DIR = '/dummy/db/dir'
    DB_FILE = '/dummy/db/file'
    DB_PATH = '/dummy/db/path'
    LOCAL_ROOT_DIR = '/dummy/local/root/dir'
    LOCAL_TMP_DIR = '/dummy/local/temp/dir'
    LOCAL_PROJ_DIR = '/dummy/local/proj/dir'
    DEFAULT_PLATFORMS_CONF = ''
    DEFAULT_JOBS_CONF = ''

class TestJobList(unittest.TestCase):
    def setUp(self):
        self.experiment_id = 'random-id'
        self.as_conf = mock.Mock()
        self.as_conf.experiment_data = dict()
        self.as_conf.experiment_data["JOBS"] = dict()
        self.as_conf.jobs_data = self.as_conf.experiment_data["JOBS"]
        self.as_conf.experiment_data["PLATFORMS"] = dict()
        self.temp_directory = tempfile.mkdtemp()
        self.JobList = JobList(self.experiment_id, FakeBasicConfig, YAMLParserFactory(),
                                JobListPersistenceDb(self.temp_directory, 'db'), self.as_conf)
        self.date_list = ["20020201", "20020202", "20020203", "20020204", "20020205", "20020206", "20020207", "20020208", "20020209", "20020210"]
        self.member_list = ["fc1", "fc2", "fc3", "fc4", "fc5", "fc6", "fc7", "fc8", "fc9", "fc10"]
        self.chunk_list = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
        self.split_list = [1, 2, 3, 4, 5]
        self.JobList._date_list = self.date_list
        self.JobList._member_list = self.member_list
        self.JobList._chunk_list = self.chunk_list
        self.JobList._split_list = self.split_list


        # Define common test case inputs here
        self.relationships_dates = {
                "DATES_FROM": {
                    "20020201": {
                        "MEMBERS_FROM": {
                            "fc2": {
                                "DATES_TO": "[20020201:20020202]*,20020203",
                                "MEMBERS_TO": "fc2",
                                "CHUNKS_TO": "all"
                            }
                        },
                        "SPLITS_FROM": {
                            "ALL": {
                                "SPLITS_TO": "1"
                            }
                        }
                    }
                }
            }
        self.relationships_dates_optional = deepcopy(self.relationships_dates)
        self.relationships_dates_optional["DATES_FROM"]["20020201"]["MEMBERS_FROM"] = { "fc2?": { "DATES_TO": "20020201", "MEMBERS_TO": "fc2", "CHUNKS_TO": "all", "SPLITS_TO": "5" } }
        self.relationships_dates_optional["DATES_FROM"]["20020201"]["SPLITS_FROM"] = { "ALL": { "SPLITS_TO": "1?" } }

        self.relationships_members = {
                "MEMBERS_FROM": {
                    "fc2": {
                        "SPLITS_FROM": {
                            "ALL": {
                                "DATES_TO": "20020201",
                                "MEMBERS_TO": "fc2",
                                "CHUNKS_TO": "all",
                                "SPLITS_TO": "1"
                            }
                        }
                    }
                }
            }
        self.relationships_chunks = {
                "CHUNKS_FROM": {
                    "1": {
                        "DATES_TO": "20020201",
                        "MEMBERS_TO": "fc2",
                        "CHUNKS_TO": "all",
                        "SPLITS_TO": "1"
                    }
                }
            }
        self.relationships_chunks2 = {
                "CHUNKS_FROM": {
                    "1": {
                        "DATES_TO": "20020201",
                        "MEMBERS_TO": "fc2",
                        "CHUNKS_TO": "all",
                        "SPLITS_TO": "1"
                    },
                    "2": {
                        "SPLITS_FROM": {
                            "5": {
                                "SPLITS_TO": "2"
                            }
                        }
                    }
                }
            }

        self.relationships_splits = {
                "SPLITS_FROM": {
                    "1": {
                        "DATES_TO": "20020201",
                        "MEMBERS_TO": "fc2",
                        "CHUNKS_TO": "all",
                        "SPLITS_TO": "1"
                    }
                }
            }

        self.relationships_general = {
                "DATES_TO": "20020201",
                "MEMBERS_TO": "fc2",
                "CHUNKS_TO": "all",
                "SPLITS_TO": "1"
            }
        self.relationships_general_1_to_1 = {
                "DATES_TO": "20020201",
                "MEMBERS_TO": "fc2",
                "CHUNKS_TO": "all",
                "SPLITS_TO": "1*,2*,3*,4*,5*"
            }
        # Create a mock Job object
        self.mock_job = mock.MagicMock(spec=Job)

        # Set the attributes on the mock object
        self.mock_job.name = "Job1"
        self.mock_job.job_id = 1
        self.mock_job.status = Status.READY
        self.mock_job.priority = 1
        self.mock_job.date = None
        self.mock_job.member = None
        self.mock_job.chunk = None
        self.mock_job.split = None

    def test_unify_to_filter(self):
        """Test the _unify_to_fitler function"""
        # :param unified_filter: Single dictionary with all filters_to
        # :param filter_to: Current dictionary that contains the filters_to
        # :param filter_type: "DATES_TO", "MEMBERS_TO", "CHUNKS_TO", "SPLITS_TO"
        # :return: unified_filter
        unified_filter = \
            {
                "DATES_TO": "20020201",
                "MEMBERS_TO": "fc2",
                "CHUNKS_TO": "all",
                "SPLITS_TO": "1"
            }
        filter_to = \
            {
                "DATES_TO": "20020205,[20020207:20020208]",
                "MEMBERS_TO": "fc2,fc3",
                "CHUNKS_TO": "all"
            }
        filter_type = "DATES_TO"
        result = self.JobList._unify_to_filter(unified_filter, filter_to, filter_type)
        expected_output = \
            {
                "DATES_TO": "20020201,20020205,20020207,20020208,",
                "MEMBERS_TO": "fc2",
                "CHUNKS_TO": "all",
                "SPLITS_TO": "1"
            }
        self.assertEqual(result, expected_output)

    def test_simple_dependency(self):
        result_d = self.JobList._check_dates({}, self.mock_job)
        result_m = self.JobList._check_members({}, self.mock_job)
        result_c = self.JobList._check_chunks({}, self.mock_job)
        result_s = self.JobList._check_splits({}, self.mock_job)
        self.assertEqual(result_d, {})
        self.assertEqual(result_m, {})
        self.assertEqual(result_c, {})
        self.assertEqual(result_s, {})

    def test_parse_filters_to_check(self):
        """Test the _parse_filters_to_check function"""
        result = self.JobList._parse_filters_to_check("20020201,20020202,20020203",self.date_list)
        expected_output = ["20020201","20020202","20020203"]
        self.assertEqual(result, expected_output)
        result = self.JobList._parse_filters_to_check("20020201,[20020203:20020205]",self.date_list)
        expected_output = ["20020201","20020203","20020204","20020205"]
        self.assertEqual(result, expected_output)
        result = self.JobList._parse_filters_to_check("[20020201:20020203],[20020205:20020207]",self.date_list)
        expected_output = ["20020201","20020202","20020203","20020205","20020206","20020207"]
        self.assertEqual(result, expected_output)
        result = self.JobList._parse_filters_to_check("20020201",self.date_list)
        expected_output = ["20020201"]
        self.assertEqual(result, expected_output)

    def test_parse_filter_to_check(self):
        # Call the function to get the result
        # Value can have the following formats:
        # a range: [0:], [:N], [0:N], [:-1], [0:N:M] ...
        # a value: N
        # a range with step: [0::M], [::2], [0::3], [::3] ...
        result = self.JobList._parse_filter_to_check("20020201",self.date_list)
        expected_output = ["20020201"]
        self.assertEqual(result, expected_output)
        result = self.JobList._parse_filter_to_check("[20020201:20020203]",self.date_list)
        expected_output = ["20020201","20020202","20020203"]
        self.assertEqual(result, expected_output)
        result = self.JobList._parse_filter_to_check("[20020201:20020203:2]",self.date_list)
        expected_output = ["20020201","20020203"]
        self.assertEqual(result, expected_output)
        result = self.JobList._parse_filter_to_check("[20020202:]",self.date_list)
        expected_output = self.date_list[1:]
        self.assertEqual(result, expected_output)
        result = self.JobList._parse_filter_to_check("[:20020203]",self.date_list)
        expected_output = self.date_list[:3]
        self.assertEqual(result, expected_output)
        result = self.JobList._parse_filter_to_check("[::2]",self.date_list)
        expected_output = self.date_list[::2]
        self.assertEqual(result, expected_output)
        result = self.JobList._parse_filter_to_check("[20020203::]",self.date_list)
        expected_output = self.date_list[2:]
        self.assertEqual(result, expected_output)
        result = self.JobList._parse_filter_to_check("[:20020203:]",self.date_list)
        expected_output = self.date_list[:3]
        self.assertEqual(result, expected_output)
        # test with a member N:N
        result = self.JobList._parse_filter_to_check("[fc2:fc3]",self.member_list)
        expected_output = ["fc2","fc3"]
        self.assertEqual(result, expected_output)
        # test with a chunk
        result = self.JobList._parse_filter_to_check("[1:2]",self.chunk_list,level_to_check="CHUNKS_FROM")
        expected_output = [1,2]
        self.assertEqual(result, expected_output)
        # test with a split
        result = self.JobList._parse_filter_to_check("[1:2]",self.split_list,level_to_check="SPLITS_FROM")
        expected_output = [1,2]
        self.assertEqual(result, expected_output)


    def test_check_dates(self):
        # Call the function to get the result
        self.mock_job.date = datetime.strptime("20020201", "%Y%m%d")
        self.mock_job.member = "fc2"
        self.mock_job.chunk = 1
        self.mock_job.split = 1
        result = self.JobList._check_dates(self.relationships_dates, self.mock_job)
        expected_output = {
                "DATES_TO": "20020201*,20020202*,20020203",
                "MEMBERS_TO": "fc2",
                "CHUNKS_TO": "all",
                "SPLITS_TO": "1"
            }
        self.assertEqual(result, expected_output)
        # failure
        self.mock_job.date = datetime.strptime("20020301", "%Y%m%d")
        result = self.JobList._check_dates(self.relationships_dates, self.mock_job)
        self.assertEqual(result, {})


    def test_check_members(self):
        # Call the function to get the result
        self.mock_job.date = datetime.strptime("20020201", "%Y%m%d")
        self.mock_job.member = "fc2"

        result = self.JobList._check_members(self.relationships_members, self.mock_job)
        expected_output = {
                "DATES_TO": "20020201",
                "MEMBERS_TO": "fc2",
                "CHUNKS_TO": "all",
                "SPLITS_TO": "1"
            }
        self.assertEqual(result, expected_output)
        self.mock_job.member = "fc3"
        result = self.JobList._check_members(self.relationships_members, self.mock_job)
        self.assertEqual(result, {})
        # FAILURE
        self.mock_job.member = "fc99"
        result = self.JobList._check_members(self.relationships_members, self.mock_job)
        self.assertEqual(result, {})


    def test_check_splits(self):
        # Call the function to get the result

        self.mock_job.split = 1
        result = self.JobList._check_splits(self.relationships_splits, self.mock_job)
        expected_output = {
                "DATES_TO": "20020201",
                "MEMBERS_TO": "fc2",
                "CHUNKS_TO": "all",
                "SPLITS_TO": "1"
            }
        self.assertEqual(result, expected_output)
        self.mock_job.split = 2
        result = self.JobList._check_splits(self.relationships_splits, self.mock_job)
        self.assertEqual(result, {})
        # failure
        self.mock_job.split = 99
        result = self.JobList._check_splits(self.relationships_splits, self.mock_job)
        self.assertEqual(result, {})

    def test_check_chunks(self):
        # Call the function to get the result

        self.mock_job.chunk = 1
        result = self.JobList._check_chunks(self.relationships_chunks, self.mock_job)
        expected_output = {
                "DATES_TO": "20020201",
                "MEMBERS_TO": "fc2",
                "CHUNKS_TO": "all",
                "SPLITS_TO": "1"
            }
        self.assertEqual(result, expected_output)
        self.mock_job.chunk = 2
        result = self.JobList._check_chunks(self.relationships_chunks, self.mock_job)
        self.assertEqual(result, {})
        # failure
        self.mock_job.chunk = 99
        result = self.JobList._check_chunks(self.relationships_chunks, self.mock_job)
        self.assertEqual(result, {})




    def test_check_general(self):
        # Call the function to get the result

        self.mock_job.date = datetime.strptime("20020201", "%Y%m%d")
        self.mock_job.member = "fc2"
        self.mock_job.chunk = 1
        self.mock_job.split = 1
        result = self.JobList._filter_current_job(self.mock_job,self.relationships_general)
        expected_output = {
                "DATES_TO": "20020201",
                "MEMBERS_TO": "fc2",
                "CHUNKS_TO": "all",
                "SPLITS_TO": "1"
            }
        self.assertEqual(result, expected_output)


    def test_valid_parent(self):

        # Call the function to get the result
        date_list = ["20020201", "20020202", "20020203", "20020204", "20020205", "20020206", "20020207", "20020208", "20020209", "20020210"]
        member_list = ["fc1", "fc2", "fc3"]
        chunk_list = [1, 2, 3]
        self.mock_job.splits = 10
        is_a_natural_relation = False
        # Filter_to values
        filter_ = {
                "DATES_TO": "20020201",
                "MEMBERS_TO": "fc2",
                "CHUNKS_TO": "all",
                "SPLITS_TO": "1"
            }
        # PArent job values
        self.mock_job.date = datetime.strptime("20020201", "%Y%m%d")
        self.mock_job.member = "fc2"
        self.mock_job.chunk = 1
        self.mock_job.split = 1
        child = copy.deepcopy(self.mock_job)
        result = self.JobList._valid_parent(self.mock_job, member_list, date_list, chunk_list, is_a_natural_relation, filter_,child)
        # it returns a tuple, the first element is the result, the second is the optional flag
        self.assertEqual(result, True)
        filter_ = {
                "DATES_TO": "20020201",
                "MEMBERS_TO": "fc2",
                "CHUNKS_TO": "all",
                "SPLITS_TO": "1?"
            }
        result = self.JobList._valid_parent(self.mock_job, member_list, date_list, chunk_list, is_a_natural_relation, filter_,child)
        self.assertEqual(result, True)
        filter_ = {
                "DATES_TO": "20020201",
                "MEMBERS_TO": "fc2",
                "CHUNKS_TO": "all",
                "SPLITS_TO": "1?"
            }
        self.mock_job.split = 2

        result = self.JobList._valid_parent(self.mock_job, member_list, date_list, chunk_list, is_a_natural_relation, filter_,child)
        self.assertEqual(result, False)
        filter_ = {
                "DATES_TO": "[20020201:20020205]",
                "MEMBERS_TO": "fc2",
                "CHUNKS_TO": "all",
                "SPLITS_TO": "1"
            }
        self.mock_job.split = 1
        result = self.JobList._valid_parent(self.mock_job, member_list, date_list, chunk_list, is_a_natural_relation, filter_,child)
        self.assertEqual(result, True)
        filter_ = {
                "DATES_TO": "[20020201:20020205]",
                "MEMBERS_TO": "fc2",
                "CHUNKS_TO": "all",
                "SPLITS_TO": "1"
            }
        self.mock_job.date = datetime.strptime("20020206", "%Y%m%d")
        result = self.JobList._valid_parent(self.mock_job, member_list, date_list, chunk_list, is_a_natural_relation, filter_,child)
        self.assertEqual(result, False)
        filter_ = {
                "DATES_TO": "[20020201:20020205]",
                "MEMBERS_TO": "fc2",
                "CHUNKS_TO": "[2:4]",
                "SPLITS_TO": "[1:5]"
            }
        self.mock_job.date = datetime.strptime("20020201", "%Y%m%d")
        self.mock_job.chunk = 2
        self.mock_job.split = 1
        result = self.JobList._valid_parent(self.mock_job, member_list, date_list, chunk_list, is_a_natural_relation, filter_,child)
        self.assertEqual(result, True)


    def test_valid_parent_1_to_1(self):
        child = copy.deepcopy(self.mock_job)
        child.splits = 6

        date_list = ["20020201", "20020202", "20020203", "20020204", "20020205", "20020206", "20020207", "20020208", "20020209", "20020210"]
        member_list = ["fc1", "fc2", "fc3"]
        chunk_list = [1, 2, 3]
        is_a_natural_relation = False

        # Test 1_to_1
        filter_ = {
                "DATES_TO": "[20020201:20020202],20020203,20020204,20020205",
                "MEMBERS_TO": "fc2",
                "CHUNKS_TO": "1,2,3,4,5,6",
                "SPLITS_TO": "1*,2*,3*,4*,5*,6"
            }
        self.mock_job.splits = 6
        self.mock_job.split = 1
        self.mock_job.date = datetime.strptime("20020204", "%Y%m%d")
        self.mock_job.chunk = 5
        child.split = 1
        self.mock_job.split = 1
        result = self.JobList._valid_parent(self.mock_job, member_list, date_list, chunk_list, is_a_natural_relation, filter_,child)
        self.assertEqual(result, True)
        child.split = 2
        result = self.JobList._valid_parent(self.mock_job, member_list, date_list, chunk_list, is_a_natural_relation, filter_,child)
        self.assertEqual(result, False)

    def test_valid_parent_1_to_n(self):
        self.mock_job.date = datetime.strptime("20020204", "%Y%m%d")
        self.mock_job.chunk = 5
        child = copy.deepcopy(self.mock_job)
        child.splits = 4
        self.mock_job.splits = 2

        date_list = ["20020201", "20020202", "20020203", "20020204", "20020205", "20020206", "20020207", "20020208", "20020209", "20020210"]
        member_list = ["fc1", "fc2", "fc3"]
        chunk_list = [1, 2, 3]
        is_a_natural_relation = False

        # Test 1_to_N
        filter_ = {
                "DATES_TO": "[20020201:20020202],20020203,20020204,20020205",
                "MEMBERS_TO": "fc2",
                "CHUNKS_TO": "1,2,3,4,5,6",
                "SPLITS_TO": "1*\\2,2*\\2"
            }
        child.split = 1
        self.mock_job.split = 1
        result = self.JobList._valid_parent(self.mock_job, member_list, date_list, chunk_list, is_a_natural_relation, filter_,child)
        self.assertEqual(result, True)
        child.split = 2
        self.mock_job.split = 1
        result = self.JobList._valid_parent(self.mock_job, member_list, date_list, chunk_list, is_a_natural_relation, filter_,child)
        self.assertEqual(result, True)
        child.split = 3
        self.mock_job.split = 1
        result = self.JobList._valid_parent(self.mock_job, member_list, date_list, chunk_list, is_a_natural_relation, filter_,child)
        self.assertEqual(result, False)
        child.split = 4
        self.mock_job.split = 1
        result = self.JobList._valid_parent(self.mock_job, member_list, date_list, chunk_list, is_a_natural_relation, filter_,child)
        self.assertEqual(result, False)

        child.split = 1
        self.mock_job.split = 2
        result = self.JobList._valid_parent(self.mock_job, member_list, date_list, chunk_list, is_a_natural_relation, filter_,child)
        self.assertEqual(result, False)
        child.split = 2
        self.mock_job.split = 2
        result = self.JobList._valid_parent(self.mock_job, member_list, date_list, chunk_list, is_a_natural_relation, filter_,child)
        self.assertEqual(result, False)
        child.split = 3
        self.mock_job.split = 2
        result = self.JobList._valid_parent(self.mock_job, member_list, date_list, chunk_list, is_a_natural_relation, filter_,child)
        self.assertEqual(result, True)
        child.split = 4
        self.mock_job.split = 2
        result = self.JobList._valid_parent(self.mock_job, member_list, date_list, chunk_list, is_a_natural_relation, filter_,child)
        self.assertEqual(result, True)

    def test_valid_parent_n_to_1(self):
        self.mock_job.date = datetime.strptime("20020204", "%Y%m%d")
        self.mock_job.chunk = 5
        child = copy.deepcopy(self.mock_job)
        child.splits = 2
        self.mock_job.splits = 4

        date_list = ["20020201", "20020202", "20020203", "20020204", "20020205", "20020206", "20020207", "20020208", "20020209", "20020210"]
        member_list = ["fc1", "fc2", "fc3"]
        chunk_list = [1, 2, 3]
        is_a_natural_relation = False

        # Test N_to_1
        filter_ = {
                "DATES_TO": "[20020201:20020202],20020203,20020204,20020205",
                "MEMBERS_TO": "fc2",
                "CHUNKS_TO": "1,2,3,4,5,6",
                "SPLITS_TO": "1*\\2,2*\\2,3*\\2,4*\\2"
            }
        child.split = 1
        self.mock_job.split = 1
        result = self.JobList._valid_parent(self.mock_job, member_list, date_list, chunk_list, is_a_natural_relation, filter_,child)
        self.assertEqual(result, True)
        child.split = 1
        self.mock_job.split = 2
        result = self.JobList._valid_parent(self.mock_job, member_list, date_list, chunk_list, is_a_natural_relation, filter_,child)
        self.assertEqual(result, True)
        child.split = 1
        self.mock_job.split = 3
        result = self.JobList._valid_parent(self.mock_job, member_list, date_list, chunk_list, is_a_natural_relation, filter_,child)
        self.assertEqual(result, False)
        child.split = 1
        self.mock_job.split = 4
        result = self.JobList._valid_parent(self.mock_job, member_list, date_list, chunk_list, is_a_natural_relation, filter_,child)
        self.assertEqual(result, False)

        child.split = 2
        self.mock_job.split = 1
        result = self.JobList._valid_parent(self.mock_job, member_list, date_list, chunk_list, is_a_natural_relation, filter_,child)
        self.assertEqual(result, False)
        child.split = 2
        self.mock_job.split = 2
        result = self.JobList._valid_parent(self.mock_job, member_list, date_list, chunk_list, is_a_natural_relation, filter_,child)
        self.assertEqual(result, False)
        child.split = 2
        self.mock_job.split = 3
        result = self.JobList._valid_parent(self.mock_job, member_list, date_list, chunk_list, is_a_natural_relation, filter_,child)
        self.assertEqual(result, True)
        child.split = 2
        self.mock_job.split = 4
        result = self.JobList._valid_parent(self.mock_job, member_list, date_list, chunk_list, is_a_natural_relation, filter_,child)
        self.assertEqual(result, True)

if __name__ == '__main__':
    unittest.main()



#!/bin/env/python
from datetime import datetime, timedelta
from .utils import timedelta2hours

class JobStat(object):
    def __init__(self, name, processors, wallclock, section, date, member, chunk):
      # type: (str, int, float, str, str, str, str) -> None
      self._name = name
      self._processors = processors
      self._wallclock = wallclock
      self.submit_time = None # type: datetime
      self.start_time = None # type: datetime        
      self.finish_time = None # type: datetime
      self.completed_queue_time = timedelta()
      self.completed_run_time = timedelta()
      self.failed_queue_time = timedelta()
      self.failed_run_time = timedelta()
      self.retrial_count = 0
      self.completed_retrial_count = 0
      self.failed_retrial_count = 0
      self.section = section
      self.date = date
      self.member = member
      self.chunk = chunk     

    def inc_retrial_count(self):
        self.retrial_count += 1

    def inc_completed_retrial_count(self):
        self.completed_retrial_count += 1

    def inc_failed_retrial_count(self):
        self.failed_retrial_count += 1

    @property
    def cpu_consumption(self):
        return timedelta2hours(self._processors * self.completed_run_time) + timedelta2hours(self._processors * self.failed_run_time)

    @property
    def failed_cpu_consumption(self):
        return timedelta2hours(self._processors * self.failed_run_time)

    @property
    def real_consumption(self):
        return timedelta2hours(self.failed_run_time + self.completed_run_time)

    @property
    def expected_real_consumption(self):
        return self._wallclock

    @property
    def expected_cpu_consumption(self):
        return self._wallclock * self._processors
    
    @property
    def name(self):
        return self._name

    def get_as_dict(self):
        return {
            "name": self._name,
            "processors": self._processors,
            "wallclock": self._wallclock,
            "completedQueueTime": timedelta2hours(self.completed_queue_time),
            "completedRunTime": timedelta2hours(self.completed_run_time),
            "failedQueueTime": timedelta2hours(self.failed_queue_time),
            "failedRunTime": timedelta2hours(self.failed_run_time),
            "cpuConsumption": self.cpu_consumption,
            "failedCpuConsumption": self.failed_cpu_consumption,
            "expectedCpuConsumption": self.expected_cpu_consumption,
            "realConsumption": self.real_consumption,
            "failedRealConsumption": timedelta2hours(self.failed_run_time),
            "expectedConsumption": self.expected_real_consumption,
            "retrialCount": self.retrial_count,
            "submittedCount": self.retrial_count,
            "completedCount": self.completed_retrial_count,
            "failedCount": self.failed_retrial_count
        }



#!/usr/bin/env python3

# Copyright 2015-2023 Earth Sciences Department, BSC-CNS

# This file is part of Autosubmit.

# Autosubmit is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Autosubmit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with Autosubmit.  If not, see <http://www.gnu.org/licenses/>.

import cProfile
import io
import os
import pstats
from datetime import datetime
from pathlib import Path
from pstats import SortKey

import psutil
from autosubmitconfigparser.config.basicconfig import BasicConfig

from log.log import Log, AutosubmitCritical


_UNITS = ["B", "KiB", "MiB", "GiB", "TiB", "PiB"]

class Profiler:
    """Class to profile the execution of experiments."""

    def __init__(self, expid:str):
        self._profiler = cProfile.Profile()
        self._expid = expid

        # Memory profiling variables
        self._mem_init = 0
        self._mem_final = 0

        # Error handling flags
        self._started = False

    def start(self) -> None:
        """Function to start the profiling process."""
        if self._started:
            raise AutosubmitCritical('The profiling process was already started.', 7074)

        self._profiler.enable()
        self._mem_init += _get_current_memory()
        self._started = True

    def stop(self) -> None:
        """Function to finish the profiling process."""
        if not self._started:
            raise AutosubmitCritical('Cannot stop the profiler because was not running.', 7074)

        self._profiler.disable()
        self._mem_final += _get_current_memory()
        self._report()

    def _report(self) -> None:
        """Function to print the final report into the stdout, log and filesystem."""

        # Create the profiler path if it does not exist
        report_path = Path(BasicConfig.LOCAL_ROOT_DIR, self._expid, "tmp", "profile")
        report_path.mkdir(parents=True, exist_ok=True)
        report_path.chmod(0o755)
        if not os.access(report_path, os.W_OK):  # Check for write access
            raise AutosubmitCritical(
                f'Directory {report_path} not writable. Please check permissions.', 7012)


        stream = io.StringIO()
        date_time = datetime.now().strftime('%Y%m%d-%H%M%S')

        # Generate function-by-function profiling results
        sort_by = SortKey.CUMULATIVE
        stats = pstats.Stats(self._profiler, stream=stream)  # generate statistics
        stats.strip_dirs().sort_stats(sort_by).print_stats() # format and save in the stream

        # Generate memory profiling results
        mem_total : float = self._mem_final - self._mem_init  # memory in Bytes
        unit = 0
        # reduces the value to its most suitable unit
        while mem_total >= 1024 and unit <= len(_UNITS) - 1:
            unit += 1
            mem_total /= 1024

        # Create and save report
        report = "\n".join([
            _generate_title("Time & Calls Profiling"),
            "",
            stream.getvalue(),
            _generate_title("Memory Profiling"),
            f"MEMORY CONSUMPTION: {mem_total} {_UNITS[unit]}.",
            ""
        ]).replace('{', '{{').replace('}', '}}') # escape {} so Log can call str.format

        Log.info(report)

        stats.dump_stats(Path(report_path, f"{self._expid}_profile_{date_time}.prof"))
        with open(Path(report_path, f"{self._expid}_profile_{date_time}.txt"),
            'w', encoding='UTF-8') as report_file:
            report_file.write(report)

        Log.info(f"[INFO] You can also find report and prof files at {report_path}\n")

def _generate_title(title="") -> str:
    """
    Generates a title banner with the specified text.

    :param title: The title that will be shown in the banner.
    :type title: str
    :return: The banner with the specified title.
    :rtype: str
    """
    max_len = 80
    separator = "=" * max_len
    message = title.center(max_len)
    return "\n".join([separator, message, separator])


def _get_current_memory() -> int:
    """
    Return the current memory consumption of the process in Bytes.

    :return: The current memory used by the process (Bytes).
    :rtype: int
    """
    return psutil.Process(os.getpid()).memory_info().rss
